import json
from pathlib import Path

try:
    from huggingface_hub import snapshot_download
except ImportError as e:
    print(
        "Error importing huggingface_hub module please run pip install huggingface_hub or upgrade the sdk "
        "by running pip install huggingface_hub --upgrade"
    )
    raise e


def snapshot_download_local(repo_id: str, local_dir: str, tokenizer_only: bool = False):
    local_dir = local_dir.rstrip("/")
    model_local_path = f"{local_dir}/{repo_id}"
    kwargs = {}
    if tokenizer_only:
        kwargs["ignore_patterns"] = ["*.bin", "*.safetensors"]
    snapshot_download(repo_id=repo_id, local_dir=model_local_path, **kwargs)
    return model_local_path


def ensure_chat_template(tokenizer_file: str, chat_template_key: str = "chat_template"):
    tokenizer_file_path = Path(tokenizer_file)
    assert (
        tokenizer_file_path.exists()
    ), f"Tokenizer config file not found at {tokenizer_file}"
    with open(tokenizer_file, "r") as f:
        tokenizer_config = json.loads(f.read())
        chat_template = tokenizer_config.get(chat_template_key)
        if chat_template is None:
            raise ValueError(
                f"Chat template not found in tokenizer config file {tokenizer_file}"
            )



# Databricks notebook source
# MAGIC %pip install mlflow-extensions
# MAGIC %pip install mlflow -U
# MAGIC %pip install sglang==0.2.13 outlines==0.0.44
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeploy
from mlflow_extensions.databricks.prebuilt import prebuilt

deployer = EzDeploy(
  config=prebuilt.vision.sglang.LLAVA_NEXT_LLAMA3_8B_CONFIG,
  registered_model_name="main.default.llava_next_llama3_8b_based"
)

deployer.download()

deployer.register()

endpoint_name = "llava_next_llama3_8b_sglang"

deployer.deploy(endpoint_name)

# COMMAND ----------

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow.utils.databricks_utils import get_databricks_host_creds

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = "llava_next_llama3_8b_sglang"
endpoint_url = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"

token = get_databricks_host_creds().token

client = OpenAI(
  base_url=endpoint_url,
  api_key=token
)
client = OpenAI(base_url=endpoint_url, api_key=token)
# print(client.models.list())
response = client.chat.completions.create(
  model="lmms-lab/llama3-llava-next-8b",
  max_tokens=256,
  messages=[
    {"role": "user", "content": [
                {"type": "text", "text": "Explain the content of the image? What is in the background?"},
                {
                    "type": "image_url",
                    "image_url": {
                      "url": "https://richmedia.ca-richimage.com/ImageDelivery/imageService?profileId=12026540&id=1859027&recipeId=728"
                    },
                },
       ],
     }
  ],
)
print(response.choices[0].message.content)

# COMMAND ----------

from mlflow_extensions.serving.compat.sglang import RuntimeEndpoint
from mlflow.utils.databricks_utils import get_databricks_host_creds
from sglang import set_default_backend
from sglang.srt.constrained import build_regex_from_object
import sglang as sgl
from pydantic import BaseModel
from typing import Literal

import requests

# the first run takes a bit longer to compile the FSM on the server, all subsequent requests will be fast
# the first call may take 10-30 seconds depending on the complexity of the pydantic object

token = get_databricks_host_creds().token

# connect sglang frontend (this python code) to the backend (model serving endpoint)
set_default_backend(RuntimeEndpoint(endpoint_url, token))


class Fashion(BaseModel):
    color: Literal["black", "blue", "gray"]
    material: Literal["silk", "denim", "fabric"]
    gender: Literal["male", "female"]

fashion = build_regex_from_object(Fashion)
# fix a small regex bug with outlines + sglang for strings
fashion = fashion.replace(r"""([^"\\\x00-\x1f\x7f-\x9f]|\\\\)""", "[\w\d\s]")
print(fashion)

@sgl.function
def image_qa(s, image_file):
    s += sgl.user(sgl.image(image_file))
    s += "Fill in the details about the item... \n"
    s += sgl.gen(
        "clothing_details",
        max_tokens=128,
        temperature=0,
        regex=fashion,  # Requires pydantic >= 2.0
    )


# URL you want to fetch
url = "https://richmedia.ca-richimage.com/ImageDelivery/imageService?profileId=12026540&id=1859027&recipeId=728"

response = requests.get(url)
response.raise_for_status()  # Check for request errors

# only need to send the bytes no download, etc
data = image_qa.run(
  image_file=response.content,
)

# access by the generation key you asked for
print(data["clothing_details"])



# COMMAND ----------

from mlflow_extensions.serving.compat.sglang import RuntimeEndpoint
from mlflow.utils.databricks_utils import get_databricks_host_creds
from sglang import set_default_backend
from sglang.srt.constrained import build_regex_from_object
import sglang as sgl
from pydantic import BaseModel
from typing import Literal

import requests

# the first run takes a bit longer to compile the FSM on the server, all subsequent requests will be fast
# the first call may take 10-30 seconds depending on the complexity of the pydantic object

token = get_databricks_host_creds().token

# connect sglang frontend (this python code) to the backend (model serving endpoint)
set_default_backend(RuntimeEndpoint(endpoint_url, token))


class FashionProblems(BaseModel):
    description: str
    clothing_type: Literal["shirt", "pants", "dress", "skirt", "shoes"]
    color: Literal["black", "blue", "gray"]
    material: Literal["silk", "denim", "fabric"]

fashion_problems = build_regex_from_object(FashionProblems)
# fix a small regex bug with outlines + sglang for strings
fashion_problems = fashion_problems.replace(r"""([^"\\\x00-\x1f\x7f-\x9f]|\\\\)""", "[\w\d\s]")
print(fashion_problems)

@sgl.function
def image_qa(s, image_file):
    s += sgl.user(sgl.image(image_file))
    s += "Fill in the problems about the product... \n"
    s += sgl.gen(
        "clothing_details",
        max_tokens=128,
        temperature=0,
        regex=fashion_problems,  # Requires pydantic >= 2.0
    )


# URL you want to fetch
url = "https://m.media-amazon.com/images/I/51a94AxNRPL.jpg"

response = requests.get(url)
response.raise_for_status()  # Check for request errors

# only need to send the bytes no download, etc
data = image_qa.run(
  image_file=response.content,
)

# access by the generation key you asked for
print(data["clothing_details"])



# COMMAND ----------

from mlflow_extensions.serving.compat.sglang import RuntimeEndpoint
from mlflow.utils.databricks_utils import get_databricks_host_creds
from sglang import set_default_backend
from sglang.srt.constrained import build_regex_from_object
import sglang as sgl
from pydantic import BaseModel
from typing import Literal, List

import requests

# the first run takes a bit longer to compile the FSM on the server, all subsequent requests will be fast
# the first call may take 10-30 seconds depending on the complexity of the pydantic object

token = get_databricks_host_creds().token

# connect sglang frontend (this python code) to the backend (model serving endpoint)
set_default_backend(RuntimeEndpoint(endpoint_url, token))


class StockoutProblems(BaseModel):
    description: str
    stockout: bool
    types_of_products: List[Literal["food", "clothing", "appliances", "durables"]]


stockout_problems = build_regex_from_object(StockoutProblems)
# fix a small regex bug with outlines + sglang for strings
stockout_problems = stockout_problems.replace(r"""([^"\\\x00-\x1f\x7f-\x9f]|\\\\)""", "[\w\d\s]")
print(stockout_problems)

@sgl.function
def image_qa(s, image_file):
    s += sgl.user(sgl.image(image_file))
    s += "Is there an out of stock situation? What type of product seems to be out of stock? (food/clothing/appliances/durables) \n"
    s += sgl.gen(
        "stockout_details",
        max_tokens=128,
        temperature=0,
        regex=stockout_problems,  # Requires pydantic >= 2.0
    )


# URL you want to fetch
url = "https://assets.eposnow.com/public/content-images/pexels-roy-broo-empty-shelves-grocery-items.jpg"

response = requests.get(url)
response.raise_for_status()  # Check for request errors

# only need to send the bytes no download, etc
data = image_qa.run(
  image_file=response.content,
)

# access by the generation key you asked for
print(data["stockout_details"])



# COMMAND ----------



import functools
import importlib
import sys
from typing import Optional

from mlflow_extensions.serving.compat import (
    CustomMLFlowHttpClient,
    inject_mlflow_openai_compat_client,
)

MODULE = "sglang"
INSTALL = "sglang"

try:
    from sglang import OpenAI
except ImportError as e:
    print(
        f"Error importing {MODULE} module please run "
        f"pip install {INSTALL} or upgrade the sdk by running pip install {INSTALL} --upgrade"
    )

OpenAI = inject_mlflow_openai_compat_client(use_sync=True)(OpenAI)


@functools.lru_cache(maxsize=32)
def get_client(url: str, api_key: str) -> "CustomMLFlowHttpClient":
    return CustomMLFlowHttpClient(
        endpoint_url=url, token=api_key, timeout=30, requires_openai_compat=False
    )


def RuntimeEndpoint(
    base_url: str,
    api_key: Optional[str] = None,
    verify: Optional[str] = None,
):
    import httpx
    from sglang import RuntimeEndpoint

    def patched_http_request(url, json=None, stream=False, api_key=None, verify=None):
        """A patched version of the httpx request to modify sglang interaction"""
        headers = {"Content-Type": "application/json; charset=utf-8"}

        parsed_url = httpx.URL(url)
        client = get_client(base_url, api_key)
        resp = client.send(
            httpx.Request(
                method="POST" if json is not None else "GET",
                json=json,
                headers=headers,
                url=parsed_url,
            )
        )
        return resp

    sys.modules["sglang.utils"].__dict__["http_request"] = patched_http_request
    importlib.reload(sys.modules["sglang.lang.backend.runtime_endpoint"])
    return RuntimeEndpoint(base_url, api_key, verify)



# Databricks notebook source
# MAGIC %pip install mlflow-extensions
# MAGIC %pip install -U mlflow
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

from mlflow_extensions.databricks.prebuilt import prebuilt
from mlflow_extensions.databricks.deploy.ez_deploy import EzDeploy

deployer = EzDeploy(
  config=prebuilt.vision.vllm.PHI_3_5_VISION_INSTRUCT_12K,
  registered_model_name="main.default.sri_phi_3_5_vision_instruct_12k"
)

# COMMAND ----------

deployer.download()

# COMMAND ----------

deployer.register()

# COMMAND ----------

endpoint_name = "sri_phi_3_5_vision_instruct_vllm"
deployer.deploy(endpoint_name)

# COMMAND ----------

dbutils.notebook.exit()

# COMMAND ----------

# MAGIC %md 
# MAGIC
# MAGIC ### Wait for model to deploy
# MAGIC

# COMMAND ----------

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow.utils.databricks_utils import get_databricks_host_creds

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"
token = get_databricks_host_creds().token

# COMMAND ----------

client = OpenAI(
  base_url=endpoint_name,
  api_key=token
)

my_model = None
for model in client.models.list():
  print(model.id)
  my_model = model.id

# COMMAND ----------

# Try this if you want to do guided decoding
# from pydantic import BaseModel

# class ExpectedJson(BaseModel):
#     outside: bool
#     inside: bool
#     boardwalk: bool
#     grass: bool

# while True:
response = client.chat.completions.create(
  model=my_model,
  messages=[
    {"role": "user", 
    "content": [
                {"type": "text", "text": "Is the image indoors or outdoors?"},
                {
                    "type": "image_url",
                    "image_url": {
                      "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
                    },
                },
            ],
    }
  ],
  # extra_body={
  #   "guided_choice": ["outside", "indoors"]
  # }
  # extra_body={
  #   "guided_json": ExpectedJson.schema()
  # }
)

response.choices[0].message.content.strip()

# COMMAND ----------





from dataclasses import dataclass, field

from mlflow_extensions.databricks.prebuilt.vision.sglang import SglangVision
from mlflow_extensions.databricks.prebuilt.vision.vllm import VllmVision


@dataclass(frozen=True)
class VisionLM:
    sglang: SglangVision = field(default_factory=lambda: SglangVision())
    vllm: VllmVision = field(default_factory=lambda: VllmVision())



from dataclasses import dataclass, field

from mlflow_extensions.databricks.prebuilt.text.sglang import SglangText
from mlflow_extensions.databricks.prebuilt.text.vllm import VllmText


@dataclass(frozen=True)
class TextLM:
    sglang: SglangText = field(default_factory=lambda: SglangText())
    vllm: VllmText = field(default_factory=lambda: VllmText())



import os
import subprocess
from dataclasses import dataclass


@dataclass
class DiagnosticCmd:
    key: str
    cmd: str
    desc: str


def run_cmd(cmd: DiagnosticCmd):
    try:
        # Run the command
        result = subprocess.run(
            cmd.cmd,
            shell=True,  # Use shell=True to allow command string input
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,  # Ensure output is returned as a string
            preexec_fn=os.setsid,  # Run in a new session
        )

        # Return results
        return {
            "command": cmd.cmd,
            "command_desc": cmd.desc,
            "stdout": result.stdout.strip(),  # Strip to remove any extra newlines
            "stderr": result.stderr.strip(),  # Strip to remove any extra newlines
            "returncode": result.returncode,
        }

    except Exception as e:
        # Return an error message
        return {
            "command": cmd.cmd,
            "command_desc": cmd.desc,
            "stdout": None,  # Strip to remove any extra newlines
            "stderr": str(e),  # Strip to remove any extra newlines
            "returncode": -1,
        }


COMPUTE_DETAILS_COMMANDS = [
    DiagnosticCmd("nvidia-smi", "nvidia-smi", "NVIDIA System Management Interface"),
    DiagnosticCmd("nvidia-topology", "nvidia-smi topo -m", "NVIDIA GPU Topology"),
    DiagnosticCmd("disk-space", "df -h", "Disk Space"),
    DiagnosticCmd("current-user", "whoami", "Current User"),
    DiagnosticCmd("active-procs", "ps aux", "Processes"),
    DiagnosticCmd("check-shm", "touch /dev/shm/test_shm", "Shared Memory Test"),
    DiagnosticCmd("pip-list", "pip list", "Installed Python Packages"),
    DiagnosticCmd("python-version", "python --version", "Python Version"),
    DiagnosticCmd("os-details", "lsb_release -a", "Linux Distribution"),
    DiagnosticCmd("arch-details", "uname -m", "Machine Architecture"),
    DiagnosticCmd("env", "env", "Environment Variables"),
    DiagnosticCmd("cpu-count", "nproc", "Number of Processors"),
    DiagnosticCmd("memory-count", "free -h", "Memory Usage"),
]

COMPUTE_DETAILS_COMMANDS_MAP = {cmd.key: cmd for cmd in COMPUTE_DETAILS_COMMANDS}


def get_compute_details(cmd_key: str) -> dict:
    if cmd_key == "all":
        return {"compute_details": [run_cmd(cmd) for cmd in COMPUTE_DETAILS_COMMANDS]}
    elif cmd_key in COMPUTE_DETAILS_COMMANDS_MAP:
        return {"compute_details": run_cmd(COMPUTE_DETAILS_COMMANDS_MAP[cmd_key])}
    else:
        return {
            "compute_details": {
                "command": cmd_key,
                "stdout": None,
                "stderr": f"Command {cmd_key} not found",
                "returncode": -1,
            }
        }



import json
import re
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import List, Literal, Optional, Type

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, ResourceDoesNotExist
from databricks.sdk.service.serving import EndpointCoreConfigInput, ServedEntityInput

from mlflow_extensions.databricks.deploy.ez_deploy_lite import EzDeployLiteManager
from mlflow_extensions.databricks.deploy.gpu_configs import (
    ALL_VALID_GPUS,
    Cloud,
    GPUConfig,
)
from mlflow_extensions.log import LogConfig
from mlflow_extensions.serving.engines.base import EngineConfig, EngineProcess
from mlflow_extensions.serving.wrapper import (
    ARCHIVE_LOG_PATH_KEY,
    ENABLE_DIAGNOSTICS_FLAG,
    LOG_FILE_KEY,
    CustomServingEnginePyfuncWrapper,
)


@dataclass(kw_only=True, frozen=True)
class ServingConfig:
    valid_gpus: List[GPUConfig] = field(default_factory=lambda: ALL_VALID_GPUS)
    minimum_memory_in_gb: Optional[int] = None

    def smallest_gpu(self, cloud: Cloud) -> GPUConfig:
        filtered_gpus = [gpu for gpu in self.valid_gpus if gpu.cloud == cloud]
        if self.minimum_memory_in_gb is not None:
            filtered_gpus = [
                gpu
                for gpu in filtered_gpus
                if gpu.total_memory_gb > self.minimum_memory_in_gb
            ]
        if len(filtered_gpus) == 0:
            raise ValueError(
                "No valid gpus in cloud please provide the right gpu config"
            )
        return min(filtered_gpus, key=lambda gpu: gpu.total_memory_gb)


@dataclass(frozen=True, kw_only=True)
class EzDeployConfig:
    name: str
    engine_config: EngineConfig
    engine_proc: Type[EngineProcess]
    serving_config: ServingConfig
    pip_config_override: List[str] = None

    def serialize_json(self):
        data = {
            "name": self.name,
            "engine_config": asdict(self.engine_config),
            "engine_proc": self.engine_proc.__name__,
            "pip_config_override": self.pip_config_override,
        }
        return json.dumps(data)

    @classmethod
    def from_json(cls, json_str):
        data = json.loads(json_str)
        engine_proc = data["engine_proc"]
        if engine_proc == "VLLMEngineProcess":
            from mlflow_extensions.serving.engines import (
                VLLMEngineConfig,
                VLLMEngineProcess,
            )

            engine_proc = VLLMEngineProcess
            engine_config = VLLMEngineConfig(**data["engine_config"])
        elif engine_proc == "SglangEngineProcess":
            from mlflow_extensions.serving.engines import (
                SglangEngineConfig,
                SglangEngineProcess,
            )

            engine_proc = SglangEngineProcess
            engine_config = SglangEngineConfig(**data["engine_config"])
        else:
            raise ValueError(f"Unsupported engine process {engine_proc}")
        return EzDeployConfig(
            name=data["name"],
            engine_config=engine_config,
            engine_proc=engine_proc,
            pip_config_override=data["pip_config_override"],
            serving_config=ServingConfig(minimum_memory_in_gb=-1),
        )

    def to_proc(self) -> EngineProcess:
        return self.engine_proc(config=self.engine_config)

    def download_artifacts(self, local_dir: str = "/local_disk0/models"):
        return self.engine_config.setup_artifacts(local_dir=local_dir)


class EzDeploy:

    def __init__(
        self,
        *,
        config: EzDeployConfig,
        registered_model_name: str,
        databricks_host: str = None,
        databricks_token: str = None,
    ):
        self._config = config
        self._registered_model_name = registered_model_name
        assert len(self._registered_model_name.split(".")) == 3, (
            "Ensure that your registered model name "
            "follows the level namespace;"
            " <catalog>.<schema>.<model_name>"
        )
        self._downloaded = False
        self._model: Optional[CustomServingEnginePyfuncWrapper] = None
        self._latest_registered_model_version = None
        if databricks_host is None or databricks_token is None:
            from mlflow.utils.databricks_utils import get_databricks_host_creds

            self._client = WorkspaceClient(
                host=get_databricks_host_creds().host,
                token=get_databricks_host_creds().token,
            )
            self._cloud = Cloud.from_host(get_databricks_host_creds().host)
        else:
            self._client = WorkspaceClient(host=databricks_host, token=databricks_token)
            self._cloud = Cloud.from_host(databricks_host)

    def download(self, *, local_dir=None):
        self._model = CustomServingEnginePyfuncWrapper(
            engine=self._config.engine_proc, engine_config=self._config.engine_config
        )
        self._model.setup(local_dir=local_dir)
        self._downloaded = True

    def register(self):
        assert self._downloaded is True, "Ensure you run the download method first"
        assert self._model is not None, "Ensure you run download"
        assert (
            self._registered_model_name is not None
        ), "Ensure you provide a valid registered_name"

        import mlflow

        mlflow.set_registry_uri("databricks-uc")

        from mlflow.models import infer_signature

        with mlflow.start_run():
            logged_model = mlflow.pyfunc.log_model(
                "model",
                python_model=self._model,
                artifacts=self._model.artifacts,
                pip_requirements=self._config.pip_config_override
                or self._model.get_pip_reqs(),
                signature=infer_signature(
                    model_input=["string-request"], model_output=["string-response"]
                ),
                registered_model_name=self._registered_model_name,
            )
            mlflow.log_params(
                {
                    "model": self._config.engine_config.model,
                    "host": self._config.engine_config.host,
                    "port": self._config.engine_config.port,
                    "command": str(self._config.engine_config.to_run_command()),
                    "serialized_engine_config": str(
                        json.dumps(asdict(self._config.engine_config))
                    ),
                }
            )
            self._latest_registered_model_version = (
                logged_model.registered_model_version
            )

    def _does_endpoint_exist(self, name) -> bool:
        try:
            self._client.serving_endpoints.get(name)
            return True
        except ResourceDoesNotExist:
            return False

    def _throw_if_volume_does_not_exist(self, path: str) -> None:
        def _path_to_volume(path: str) -> str:
            pattern: str = r"^(?:dbfs:)?/Volumes/([^/]+)/([^/]+)/([^/]+)/?.*"
            match: re.Match = re.match(pattern, path)
            if match:
                catalog: str = match.group(1)
                schema: str = match.group(2)
                volume: str = match.group(3)
                return f"{catalog}.{schema}.{volume}"
            raise ValueError(f"Invalid path {path}")

        try:
            volume: str = _path_to_volume(path)
            self._client.volumes.read(volume)
        except NotFound:
            raise ValueError(f"Volume {volume} does not exist")

    def deploy(
        self,
        name,
        *,
        scale_to_zero: bool = True,
        workload_size: Literal["Small", "Medium", "Large"] = "Small",
        workload_type: Optional[str] = None,
        enable_diagnostics: bool = False,
        log_config: Optional[LogConfig] = None,
    ):
        gpu_cfg = self._config.serving_config.smallest_gpu(self._cloud)
        endpoint_exists = self._does_endpoint_exist(name)
        try:
            from pyspark.sql import SparkSession

            spark = SparkSession.getActiveSession()
            workspace_url = spark.conf.get("spark.databricks.workspaceUrl")
            print(
                f"Deploying model: {name} track progress here: https://{workspace_url}/ml/endpoints/{name}"
            )
        except ImportError:
            print(
                f"Deploying model: {name}; look at the serving tab to track progress..."
            )
        environment_vars = {
            # ensure that for docker container that all logs are always flushed to terminal
            # and not buffered anywhere
            "PYTHONUNBUFFERED": "1",
        }
        if enable_diagnostics is True:
            environment_vars[ENABLE_DIAGNOSTICS_FLAG] = "true"
        if log_config is not None:
            log_file: str = log_config.filename or f"{name}.log"
            environment_vars[LOG_FILE_KEY] = log_file
            if log_config.archive_path is not None:
                self._throw_if_volume_does_not_exist(log_config.archive_path)
                archive_log_path: Path = Path(log_config.archive_path)
                archive_log_path = (
                    archive_log_path
                    if archive_log_path.name == name
                    else archive_log_path / name
                )
                environment_vars[ARCHIVE_LOG_PATH_KEY] = archive_log_path.as_posix()
        if endpoint_exists is False:
            self._client.serving_endpoints.create(
                name=name,
                config=EndpointCoreConfigInput(
                    name=name,
                    served_entities=[
                        ServedEntityInput(
                            name=name,
                            entity_name=self._registered_model_name,
                            entity_version=self._latest_registered_model_version,
                            scale_to_zero_enabled=scale_to_zero,
                            workload_type=workload_type or gpu_cfg.name,
                            workload_size=workload_size,
                            environment_vars=environment_vars,
                        )
                    ],
                ),
            )
        else:
            self._client.serving_endpoints.update_config(
                name=name,
                served_entities=[
                    ServedEntityInput(
                        name=name,
                        entity_name=self._registered_model_name,
                        entity_version=self._latest_registered_model_version,
                        scale_to_zero_enabled=scale_to_zero,
                        workload_type=workload_type or gpu_cfg.name,
                        workload_size=workload_size,
                        environment_vars=environment_vars,
                    )
                ],
            )


class EzDeployLite:

    def __init__(
        self,
        ez_deploy_config: EzDeployConfig,
        databricks_host: str = None,
        databricks_token: str = None,
    ):
        self._config: EzDeployConfig = ez_deploy_config
        self._downloaded = False
        if databricks_host is None or databricks_token is None:
            from mlflow.utils.databricks_utils import get_databricks_host_creds

            self._client = WorkspaceClient(
                host=get_databricks_host_creds().host,
                token=get_databricks_host_creds().token,
            )
            self._cloud = Cloud.from_host(get_databricks_host_creds().host)
        else:
            self._client = WorkspaceClient(host=databricks_host, token=databricks_token)
            self._cloud = Cloud.from_host(databricks_host)
        self._edlm = EzDeployLiteManager(
            databricks_host=databricks_host, databricks_token=databricks_token
        )

    def deploy(
        self,
        deployment_name: str,
        hf_secret_scope: str,
        hf_secret_key: str,
        specific_git_ref: str = None,
    ):
        self._edlm.upsert(
            deployment_name,
            cloud_provider=self._cloud,
            ez_deploy_config=self._config,
            hf_secret_key=hf_secret_key,
            hf_secret_scope=hf_secret_scope,
            entrypoint_git_ref=specific_git_ref,
        )
        self._edlm.start_server(deployment_name)



from dataclasses import dataclass
from enum import Enum


class Cloud(Enum):
    AWS = "AWS"
    AZURE = "AZURE"
    GCP = "GCP"

    @staticmethod
    def from_host(host):
        if "gcp.databricks.com" in host:
            return Cloud.GCP
        if "azuredatabricks.net" in host:
            return Cloud.AZURE
        return Cloud.AWS


class GPUType(Enum):
    T4 = "T4"
    A10G = "A10G"
    A100_80G = "A100_80G"
    A100_40G = "A100_40G"
    H100 = "H100"

    @property
    def memory_gb(self):
        memory_map = {
            "T4": 16,
            "A10G": 24,
            "A100_80G": 80,
            "A100_40G": 40,
            "H100": 80,
        }
        return memory_map.get(self.value, "Unknown")


@dataclass(frozen=True, kw_only=True)
class GPUConfig:
    name: str
    gpu_count: int
    gpu_type: GPUType
    cloud: Cloud

    def __repr__(self):
        return f"{self.name}({self.gpu_count}x{self.gpu_type})"

    @property
    def total_memory_gb(self):
        return self.gpu_count * self.gpu_type.memory_gb

    @property
    def single_gpu_memory_gb(self):
        return self.gpu_type.memory_gb


class AWSServingGPUConfig(Enum):
    GPU_SMALL = GPUConfig(
        name="GPU_SMALL", gpu_count=1, gpu_type=GPUType.T4, cloud=Cloud.AWS
    )  # 1xT4
    GPU_MEDIUM = GPUConfig(
        name="GPU_MEDIUM", gpu_count=1, gpu_type=GPUType.A10G, cloud=Cloud.AWS
    )  # 1xA10G
    MULTIGPU_MEDIUM = GPUConfig(
        name="MULTIGPU_MEDIUM", gpu_count=4, gpu_type=GPUType.A10G, cloud=Cloud.AWS
    )  # 4xA10G
    GPU_MEDIUM_8 = GPUConfig(
        name="GPU_MEDIUM_8", gpu_count=8, gpu_type=GPUType.A10G, cloud=Cloud.AWS
    )  # 8xA10G
    GPU_LARGE_8 = GPUConfig(
        name="GPU_LARGE_8", gpu_count=8, gpu_type=GPUType.A100_80G, cloud=Cloud.AWS
    )  # 8xA100_80G
    GPU_XLARGE_8 = GPUConfig(
        name="GPU_XLARGE_8", gpu_count=8, gpu_type=GPUType.H100, cloud=Cloud.AWS
    )  # 8xH100


class AzureServingGPUConfig(Enum):
    GPU_SMALL = GPUConfig(
        name="GPU_SMALL", gpu_count=1, gpu_type=GPUType.T4, cloud=Cloud.AZURE
    )  # 1xT4
    GPU_LARGE = GPUConfig(
        name="GPU_LARGE", gpu_count=1, gpu_type=GPUType.A100_80G, cloud=Cloud.AZURE
    )  # 1xA100_80G
    GPU_LARGE_2 = GPUConfig(
        name="GPU_LARGE_2", gpu_count=2, gpu_type=GPUType.A100_80G, cloud=Cloud.AZURE
    )  # 2xA100_80G
    GPU_LARGE_8 = GPUConfig(
        name="GPU_LARGE_8", gpu_count=8, gpu_type=GPUType.A100_80G, cloud=Cloud.AZURE
    )  # 8xA100_80G


ALL_VALID_GPUS = [gpu.value for gpu in AzureServingGPUConfig] + [
    gpu.value for gpu in AWSServingGPUConfig
]


class AwsVMConfigs(Enum):
    G5_16XLARGE = GPUConfig(
        name="g5.16xlarge", gpu_count=1, gpu_type=GPUType.A10G, cloud=Cloud.AWS
    )
    G5_24XLARGE = GPUConfig(
        name="g5.24xlarge", gpu_count=4, gpu_type=GPUType.A10G, cloud=Cloud.AWS
    )
    G5_48XLARGE = GPUConfig(
        name="g5.48xlarge", gpu_count=8, gpu_type=GPUType.A10G, cloud=Cloud.AWS
    )
    P4D_24XLARGE = GPUConfig(
        name="p4d.24xlarge", gpu_count=8, gpu_type=GPUType.A100_80G, cloud=Cloud.AWS
    )
    P4DE_24XLARGE = GPUConfig(
        name="p4de.24xlarge", gpu_count=8, gpu_type=GPUType.A100_80G, cloud=Cloud.AWS
    )


class AzureVMConfigs(Enum):
    NV36ADS_A10_V5 = GPUConfig(
        name="Standard_NV36ads_A10_v5",
        gpu_count=1,
        gpu_type=GPUType.A10G,
        cloud=Cloud.AZURE,
    )
    NV72ADS_A10_V5 = GPUConfig(
        name="Standard_NV72ads_A10_v5",
        gpu_count=2,
        gpu_type=GPUType.A10G,
        cloud=Cloud.AZURE,
    )
    NC24ADS_A100_V4 = GPUConfig(
        name="Standard_NC24ads_A100_v4",
        gpu_count=1,
        gpu_type=GPUType.A100_80G,
        cloud=Cloud.AZURE,
    )
    NC48ADS_A100_V4 = GPUConfig(
        name="Standard_NC48ads_A100_v4",
        gpu_count=2,
        gpu_type=GPUType.A100_80G,
        cloud=Cloud.AZURE,
    )
    NC96ADS_A100_V4 = GPUConfig(
        name="Standard_NC96ads_A100_v4",
        gpu_count=4,
        gpu_type=GPUType.A100_80G,
        cloud=Cloud.AZURE,
    )


class GCPVMConfigs(Enum):
    A2_HIGHGPU_1G = GPUConfig(
        name="a2-highgpu-1g", gpu_count=1, gpu_type=GPUType.A100_40G, cloud=Cloud.GCP
    )
    A2_HIGHGPU_2G = GPUConfig(
        name="a2-highgpu-2g", gpu_count=2, gpu_type=GPUType.A100_40G, cloud=Cloud.GCP
    )
    A2_HIGHGPU_4G = GPUConfig(
        name="a2-highgpu-4g", gpu_count=4, gpu_type=GPUType.A100_40G, cloud=Cloud.GCP
    )
    A2_HIGHGPU_8G = GPUConfig(
        name="a2-highgpu-8g", gpu_count=8, gpu_type=GPUType.A100_40G, cloud=Cloud.GCP
    )
    A2_MEGAGPU_16G = GPUConfig(
        name="a2-megagpu-16g", gpu_count=16, gpu_type=GPUType.A100_40G, cloud=Cloud.GCP
    )


ALL_VALID_VM_CONFIGS = (
    [gpu.value for gpu in GCPVMConfigs]
    + [gpu.value for gpu in AzureVMConfigs]
    + [gpu.value for gpu in AwsVMConfigs]
)



import logging
import os
import sys
from typing import Callable, List

import pytest
import structlog

from mlflow_extensions.log import (
    LogConfig,
    Logger,
    LogLevel,
    get_logger,
    initialize_logging,
)
from mlflow_extensions.log.handlers import (
    FileRotator,
    VolumeRotator,
    _get_databricks_host_creds,
    create_rotator,
    full_volume_name_to_path,
    rotating_volume_handler,
)


def test_full_volume_name_to_path() -> None:
    full_volume_name: str = "foo.bar.baz"
    expected_path: str = "/Volumes/foo/bar/baz"
    assert full_volume_name_to_path(full_volume_name) == expected_path


def test_intialize_logging_with_defaults() -> None:
    initialize_logging(LogConfig(filename="test.log"))
    assert structlog.is_configured()


def test_get_basic_logger_should_be_filtered() -> None:
    initialize_logging(
        LogConfig(
            level=LogLevel.DEBUG,
            filename="/tmp/test.log",
        )
    )
    logger: Logger = logging.getLogger()
    logger.info("This is the message")


def test_get_logger_should_return_default_log_level_and_name() -> None:
    initialize_logging(
        LogConfig(
            level=LogLevel.DEBUG,
            filename="/tmp/test.log",
            archive_path="/Volumes/nfleming/main/logs/test",
        )
    )
    logger: Logger = get_logger()
    logger.info("This is the message", foo="bar", baz="qux")


def test_get_logger_should_return_default_log_level_and_name_no_volume() -> None:
    initialize_logging(LogConfig(level=LogLevel.DEBUG, filename="/tmp/test.log"))
    logger: Logger = get_logger()
    logger.info("This is the message", foo="bar", baz="qux")


def test_create_rotator_with_no_volume() -> None:
    assert isinstance(create_rotator(None, None, None), FileRotator)


@pytest.mark.skipif(
    "DATABRICKS_HOST" not in os.environ,
    reason="Missing environment variable: DATABRICKS_HOST",
)
@pytest.mark.skipif(
    "DATABRICKS_TOKEN" not in os.environ,
    reason="Missing environment variable: DATABRICKS_TOKEN",
)
def test_create_rotator_with_volume_path() -> None:
    assert isinstance(
        create_rotator("/Volumes/nfleming/default/logs", None, None), VolumeRotator
    )


@pytest.mark.skipif(
    "DATABRICKS_HOST" not in os.environ,
    reason="Missing environment variable: DATABRICKS_HOST",
)
@pytest.mark.skipif(
    "DATABRICKS_TOKEN" not in os.environ,
    reason="Missing environment variable: DATABRICKS_TOKEN",
)
def test_databricks_host_creds() -> None:
    host, token = _get_databricks_host_creds(None, None)
    assert host is not None
    assert token is not None



# Databricks notebook source
# MAGIC %md
# MAGIC
# MAGIC # These are sanity tests for ezdeploy for vLLM on A100

# COMMAND ----------

# MAGIC %pip install vllm==0.6.0 filelock httpx hf_transfer
# MAGIC %pip install -U openai
# MAGIC %pip install -U mlflow
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

# MAGIC %pip list | grep vllm

# COMMAND ----------

from mlflow_extensions.databricks.deploy.gpu_configs import AzureServingGPUConfig
from mlflow_extensions.testing.runner import (
    RequestResult,
    ServerFramework,
    run_all_tests,
)

THIS_GPU = AzureServingGPUConfig.GPU_LARGE.value
THIS_FRAMEWORK = ServerFramework.VLLM

# COMMAND ----------

import os

os.environ["HF_TOKEN"] = dbutils.secrets.get(
    scope="sri-mlflow-extensions", key="hf-token"
)
os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"

# COMMAND ----------

results = run_all_tests(gpu_config=THIS_GPU, server_framework=THIS_FRAMEWORK)

# COMMAND ----------

import pandas as pd

display(pd.DataFrame(RequestResult.make_df_friendly(results)))

# COMMAND ----------

# MAGIC %md
# MAGIC # FAILING TESTS

# COMMAND ----------

import pandas as pd

# errored records
errored_results = [result for result in results if result.is_error is True]
if len(errored_results) > 0:
    display(pd.DataFrame(RequestResult.make_df_friendly(errored_results)))
else:
    print("No tests failed...")

# COMMAND ----------

# MAGIC %md
# MAGIC # Success Records

# COMMAND ----------

import pandas as pd

# errored records
success_results = [result for result in results if result.is_error is False]
if len(success_results) > 0:
    display(pd.DataFrame(RequestResult.make_df_friendly(success_results)))
else:
    print("No tests passed...")

# COMMAND ----------

assert len(errored_results) == 0, "Tests failed"

# COMMAND ----------



import base64
import typing

import requests

from mlflow_extensions.testing.helper import Modality, inject_openai_client, run_if

if typing.TYPE_CHECKING is True:
    from openai import OpenAi

    from mlflow_extensions.testing.runner import ModelContextRunner


def encode_audio_base64_from_url(audio_url: str) -> str:
    """Encode an audio retrieved from a remote url to base64 format."""

    with requests.get(audio_url) as response:
        response.raise_for_status()
        result = base64.b64encode(response.content).decode("utf-8")

    return result


@run_if(modality=Modality.AUDIO.value)
@inject_openai_client
def query_audio(
    *,
    ctx: "ModelContextRunner",
    client: "OpenAi",
    audio_data: str,
    model: str,
    modality_type: str = None,
    host: str = "0.0.0.0",  # noqa
    port: int = 9989,  # noqa
    repeat_n: int = 5,
):
    count = 0
    # repeat a few times
    while count < repeat_n:
        try:
            response = client.chat.completions.create(
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": "Breakdown the content of the audio?",
                            },
                            {
                                "type": "audio_url",
                                "audio_url": {
                                    # Any format supported by librosa is supported
                                    "url": f"data:audio/ogg;base64,{audio_data}"
                                },
                            },
                        ],
                    }
                ],
                model=model,
                max_tokens=512,
            )
            ctx.add_success(result=response.choices[0].message.content)

            from typing import List, Literal

            from pydantic import BaseModel

            class AudioExtraction(BaseModel):
                year: str
                speaker: str
                location: str
                sentiment: Literal["positive", "negative", "neutral"]
                tone: Literal["somber", "upbeat", "pessemistic"]
                summary: str

            response = client.chat.completions.create(
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": f"Extract the following content from the audio: {str(AudioExtraction.schema())}?",
                            },
                            {
                                "type": "audio_url",
                                "audio_url": {
                                    # Any format supported by librosa is supported
                                    "url": f"data:audio/ogg;base64,{audio_data}"
                                },
                            },
                        ],
                    }
                ],
                model=model,
                max_tokens=512,
                extra_body={"guided_json": AudioExtraction.schema()},
            )
            ctx.add_success(result=response.choices[0].message.content)
        except Exception as e:
            print(e)
            ctx.add_error(error_msg=str(e))
        finally:
            count += 1



import abc
import json
import os
import random
import signal
import socket
import subprocess
import time
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from threading import Thread
from typing import Dict, List, Optional, Union

import httpx
import psutil
from filelock import FileLock
from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.log import Logger, get_logger, log_around
from mlflow_extensions.serving.engines.gpu_utils import not_enough_shm
from mlflow_extensions.version import get_mlflow_extensions_version

LOGGER: Logger = get_logger()


@log_around(logger=LOGGER)
def is_port_open(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        result = sock.connect_ex((host, port))
        return result == 0


@dataclass(kw_only=True)
class Command:
    name: str
    command: List[str]
    active_process: Optional[subprocess.Popen] = None
    long_living: bool = True
    env: Optional[Dict[str, str]] = None

    @log_around(logger=LOGGER)
    def start(self):
        if self.long_living is True:
            self.active_process = subprocess.Popen(
                self.command,
                env=self.env or os.environ.copy(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                # ensure process is in another process group / session
                preexec_fn=os.setsid,
            )
        else:
            self.active_process = subprocess.Popen(
                self.command,
                env=self.env or os.environ.copy(),
                preexec_fn=os.setsid,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,  # This will handle the output as text rather than bytes
                bufsize=1,  # Line-buffered
            )

    @log_around(logger=LOGGER)
    def is_running(self) -> bool:
        if self.active_process is None:
            return False
        if self.active_process.returncode is not None:
            return False
        return True

    @log_around(logger=LOGGER)
    def wait_and_log(self):
        if self.active_process is None:
            LOGGER.info("Process has not been started.")
            return

        if self.long_living is True:
            raise ValueError(
                "Unable to wait and log for long living process will hang the thread."
            )

        try:
            # Stream and print stdout
            for line in self.active_process.stdout:
                LOGGER.info(f"STDOUT: {line.strip()}")

            # Wait for the process to complete
            self.active_process.wait()

            # Check the exit code
            LOGGER.info("Exit Code:", self.active_process.returncode)

        except Exception as e:
            LOGGER.error(f"Error while waiting for logs: {e}", error=e)

    @log_around(logger=LOGGER)
    def stop(self):
        if self.is_running() is False:
            return
        # Send SIGTERM to the subprocess
        os.killpg(os.getpgid(self.active_process.pid), signal.SIGTERM)

        # Wait for a short while and check if the process has terminated
        time.sleep(5)
        if self.active_process.poll() is None:
            # If still running, send SIGKILL
            os.killpg(os.getpgid(self.active_process.pid), signal.SIGKILL)

        # # Optionally, wait for the process to terminate and get its exit status
        try:
            # Wait for the process to terminate, with a timeout
            stdout, stderr = self.active_process.communicate(timeout=10)
        except subprocess.TimeoutExpired:
            stdout, stderr = self.active_process.communicate()
            LOGGER.warning("Timed out")

        try:
            LOGGER.info("STDOUT:", stdout.decode())
            LOGGER.info("STDERR:", stderr.decode())
        finally:
            LOGGER.info("Exit Code:", self.active_process.returncode)


@dataclass(frozen=True, kw_only=True)
class EngineConfig(abc.ABC):
    model: str
    host: str = field(default="0.0.0.0")
    port: int = field(default=9989)
    openai_api_path: int = field(default="v1")
    ensure_supported_models: bool = field(default=True)
    library_overrides: dict = field(default_factory=dict)

    @abc.abstractmethod
    def _to_run_command(
        self, context: PythonModelContext = None
    ) -> Union[List[str], Command]:
        pass

    @log_around(logger=LOGGER)
    def to_run_command(
        self, context: PythonModelContext = None
    ) -> Union[List[str], Command]:
        command = self._to_run_command(context)
        LOGGER.info(f"Command: {command}")
        if isinstance(command, list):
            # ensure all items are strings
            return [str(item) for item in command]
        return command

    @log_around(logger=LOGGER)
    def default_pip_reqs(
        self,
        *,
        filelock_version: str = "3.15.4",
        httpx_version: str = "0.27.0",
        psutil_version: str = "6.0.0",
        mlflow_version: str = "2.16.0",
        mlflow_extensions_version: str = None,
        **kwargs,
    ) -> List[str]:
        mlflow_extensions_version = (
            mlflow_extensions_version or get_mlflow_extensions_version()
        )
        default_installs = {
            "httpx": f"httpx=={httpx_version}",
            "psutil": f"psutil=={psutil_version}",
            "filelock": f"filelock=={filelock_version}",
            "mlflow": f"mlflow=={mlflow_version}",
            "mlflow-extensions": (
                f"mlflow-extensions=={mlflow_extensions_version}"
                if mlflow_extensions_version
                else "mlflow-extensions"
            ),
        }
        final_installs = {}
        for key, value in default_installs.items():
            final_installs[key] = value
        final_installs.update(self.engine_pip_reqs(**kwargs))
        if self.library_overrides:
            final_installs.update(self.library_overrides)
        return list(final_installs.values())

    @abc.abstractmethod
    def engine_pip_reqs(self, **kwargs) -> Dict[str, str]:
        # key should be eggname/packagename and the value should be the full
        # installation string in requirements.txt
        pass

    @abc.abstractmethod
    def setup_artifacts(self, local_dir: str = "/root/models") -> Dict[str, str]:
        pass

    @staticmethod
    @abc.abstractmethod
    def supported_model_architectures(self) -> List[str]:
        pass


class EngineHealthCheckStatusManager:

    def __init__(
        self,
        health_check_path: str = "~/.mlflow-extensions/health-check.txt",
        availability_path: str = "~/.mlflow-extensions/availability.txt",
        heartbeat_path: str = "~/.mlflow-extensions/heartbeat.txt",
    ):
        self._health_check_relative_path = Path(health_check_path)
        self._availability_relative_path = Path(availability_path)
        self._heartbeat_relative_path = Path(heartbeat_path)
        self._health_check_path = (
            self._health_check_relative_path.expanduser().resolve()
        )
        self._availability_path = (
            self._availability_relative_path.expanduser().resolve()
        )
        self._heartbeat_path = self._heartbeat_relative_path.expanduser().resolve()
        self._ensure_file_exists()

    def _ensure_file_exists(self):
        # ensure parent directories
        self._health_check_path.parent.mkdir(parents=True, exist_ok=True)
        self._availability_path.parent.mkdir(parents=True, exist_ok=True)
        self._heartbeat_path.parent.mkdir(parents=True, exist_ok=True)
        if not self._health_check_path.exists():
            self._health_check_path.touch()
        if not self._availability_path.exists():
            self._availability_path.touch()
        if not self._heartbeat_path.exists():
            self._heartbeat_path.touch()

    def start_empty(self):
        with open(self._health_check_path, "w") as f:
            f.write("")

    def add_status(self, status: str):
        status = {"datetime_utc": str(datetime.now()), "status": status}
        json_status = json.dumps(status)
        with open(self._health_check_path, "a") as f:
            f.write(f"{json_status}\n")

    def set_available(self):
        with open(self._availability_path, "w") as f:
            f.write("AVAILABLE")

    def set_unavailable(self):
        with open(self._availability_path, "w") as f:
            f.write("UNAVAILABLE")

    def get_availability(self):
        with open(self._availability_path, "r") as f:
            return f.read()

    def get_last_heartbeat(self):
        with open(self._heartbeat_path, "r") as f:
            return f.read()

    def set_heartbeat(self):
        with open(self._heartbeat_path, "w") as f:
            f.write(str(datetime.now()))

    def get_last_n_status(self, n: int = 100) -> List[Dict[str, str]]:
        with open(self._health_check_path, "r") as f:
            lines = f.readlines()
        if len(lines) < n:
            return [json.loads(line) for line in lines]
        return [json.loads(line) for line in lines[-n:]]


class EngineProcess(abc.ABC):

    def __init__(self, *, config: EngineConfig):
        self._config = config
        # expand ~ to home directory
        file_lock_path = os.path.expanduser(f"~/{self.__class__.__name__}.txt.lock")
        self._lock = FileLock(file_lock_path)
        self._server_http_client = httpx.Client(
            base_url=f"http://{self._config.host}:{self._config.port}", timeout=30
        )
        self._oai_http_client = httpx.Client(
            base_url=f"http://{self._config.host}:{self._config.port}/{self._config.openai_api_path}",
            timeout=300,
        )
        self._proc = None
        self._run_health_check = None
        self._health_check_thread = None
        self._health_check_status_file = EngineHealthCheckStatusManager()

    @property
    def server_process(self):
        return self._proc

    @property
    def oai_http_client(self) -> httpx.Client:
        return self._oai_http_client

    @property
    def server_http_client(self) -> httpx.Client:
        return self._server_http_client

    @property
    @abc.abstractmethod
    def engine_name(self) -> str:
        pass

    @property
    def config(self) -> EngineConfig:
        return self._config

    @abc.abstractmethod
    def health_check(self) -> bool:
        pass

    @log_around(logger=LOGGER)
    def is_process_healthy(self) -> bool:
        if not is_port_open(self.config.host, self.config.port):
            return False

        return self.health_check()

    @log_around(logger=LOGGER)
    def stop_proc(self):
        with self._lock:
            self._kill_active_proc()
            self._proc = None
            self._run_health_check = False
            self.cleanup()

    @log_around(logger=LOGGER)
    def _is_process_running(self):
        if self._proc is None:
            return False
        try:
            process = psutil.Process(self._proc.pid)
            return process.is_running() and process.status() != psutil.STATUS_ZOMBIE
        except psutil.NoSuchProcess:
            return False

    @log_around(logger=LOGGER)
    def _kill_active_proc(self):
        os.killpg(os.getpgid(self._proc.pid), signal.SIGTERM)
        time.sleep(5)
        if self._proc.poll() is None:
            os.killpg(os.getpgid(self._proc.pid), signal.SIGKILL)

    def cleanup(self) -> None:
        # should not throw or raise errors
        pass

    @log_around(logger=LOGGER)
    def _spawn_server_proc(self, context: PythonModelContext = None):
        proc_env = os.environ.copy()
        server_details = {"HOST": self.config.host, "PORT": str(self.config.port)}
        if not_enough_shm() is True:
            LOGGER.warning(
                "Not enough shared memory for NCCL. Setting NCCL_SHM_DISABLE=1"
            )
            server_details["NCCL_SHM_DISABLE"] = "1"
        proc_env.update(server_details)
        command = self.config.to_run_command(context)
        if isinstance(command, list):
            self._proc = subprocess.Popen(
                command,
                env=proc_env,
                preexec_fn=os.setsid,
            )
        elif isinstance(command, Command):
            command.start()

        while self.is_process_healthy() is False and self._is_process_running() is True:
            LOGGER.info(f"Waiting for {self.engine_name} to start")
            time.sleep(1)

    @log_around(logger=LOGGER)
    def ensure_server_is_running(
        self,
        *,
        context: PythonModelContext = None,
        health_check_frequency_seconds: int = 10,
        max_respawn_attempts: int = 3,
    ):
        LOGGER.info(f"Ensuring {self.engine_name} is running for pid {self._proc.pid}")
        if self._proc is None:
            raise ValueError("Process not started yet. Run start_proc() first.")
        attempt_count = 0
        # check if pid is still running
        self._health_check_status_file.start_empty()
        while True:
            # let's set heartbeat before processing anything
            self._health_check_status_file.set_heartbeat()
            if attempt_count > max_respawn_attempts:
                LOGGER.info(
                    f"Max respawn attempts reached for {self.engine_name}. Restart serving endpoint."
                )
                self._health_check_status_file.add_status(
                    "Max respawn attempts reached, server unable to be respawned."
                )
                break
            if self._run_health_check is not True:
                break
            try:
                if self._is_process_running():
                    time.sleep(health_check_frequency_seconds)
                else:
                    LOGGER.info(f"Process: {self._proc.pid} is not running.")
                    self._health_check_status_file.set_unavailable()
                    self._health_check_status_file.add_status("Process is not running.")
                    process = psutil.Process(self._proc.pid)
                    if process.status() == psutil.STATUS_ZOMBIE:
                        LOGGER.info(
                            f"Process: {self._proc.pid} is zombie. Killing process."
                        )
                        self._health_check_status_file.add_status(
                            "Process is a zombie. Killing process."
                        )
                        self._kill_active_proc()
                    LOGGER.info(f"Calling cleanup for {self.engine_name}")
                    self._health_check_status_file.add_status(
                        "Process is not running. Calling cleanup."
                    )
                    self.cleanup()
                    LOGGER.info(f"Respawning {self.engine_name}")
                    self._health_check_status_file.add_status(
                        "Process is not running. Respawning."
                    )
                    self._spawn_server_proc(context)
                    self._health_check_status_file.add_status(
                        "Spawn Server Proc Finished."
                    )
                    if self.health_check() is False:
                        self._health_check_status_file.add_status(
                            "Health check failed after respawn. "
                            f"Attempt: {attempt_count}"
                        )
                        LOGGER.error(f"Health check failed after respawn.")
                        attempt_count += 1
                    else:
                        LOGGER.info(f"Health check passed after respawn.")
                        self._health_check_status_file.add_status(
                            "Health check passed after respawn."
                        )
                        self._health_check_status_file.set_available()
                        attempt_count = 0
            except psutil.NoSuchProcess:
                LOGGER.info(f"Process: {self._proc.pid} does not exist.")
                self._health_check_status_file.add_status(
                    "Process does not exist. Respawning."
                )
                LOGGER.info(f"Respawning {self.engine_name}")
                self._spawn_server_proc(context)
                LOGGER.info(f"Spawn Server Proc Finished.")
                self._health_check_status_file.add_status("Spawn Server Proc Finished.")
                if self.health_check() is False:
                    self._health_check_status_file.add_status(
                        "Health check failed after respawn. "
                        f"Attempt: {attempt_count}"
                    )
                    LOGGER.error("Health check failed after respawn.")
                    attempt_count += 1
                else:
                    self._health_check_status_file.set_available()
                    self._health_check_status_file.add_status(
                        "Health check passed after respawn."
                    )
                    LOGGER.info("Health check passed after respawn.")
                    attempt_count = 0

    def health_check_status(self):
        return {
            "status": self._health_check_status_file.get_availability(),
            "worker_pid": os.getpid(),
            "engine_name": self.engine_name,
            "health_check_last_heartbeat": self._health_check_status_file.get_last_heartbeat(),
            "health_check_thread_last_50_status": self._health_check_status_file.get_last_n_status(
                50
            ),
            "note": "heartbeat is updated at a frequency but can pause when server is being respawned",
        }

    @log_around(logger=LOGGER)
    # todo add local lora paths
    def start_proc(
        self, context: PythonModelContext = None, health_check_thread: bool = True
    ):
        # kill process in port if already running
        time.sleep(random.randint(1, 5))
        LOGGER.info(f"Attempting to acquire Lock")
        with self._lock:
            LOGGER.info(f"Acquired Lock")
            if self.health_check() is False:
                self._health_check_status_file.set_unavailable()
                self._spawn_server_proc(context)
                self._health_check_status_file.set_available()
                self._run_health_check = True
                if health_check_thread is True:
                    self._health_check_thread = Thread(
                        target=self.ensure_server_is_running,
                        kwargs={
                            "context": context,
                        },
                    )
                    if self._health_check_thread is not None:
                        self._health_check_thread.start()
            else:
                LOGGER.info(f"{self.engine_name} already running")



import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, List, Optional, Type

import mlflow
import numpy as np
import pandas as pd
from httpx import Request, Response
from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.log import LogConfig, Logger, get_logger, initialize_logging
from mlflow_extensions.serving.compute_details import get_compute_details
from mlflow_extensions.serving.engines.base import EngineConfig, EngineProcess
from mlflow_extensions.serving.serde import ResponseMessageV1
from mlflow_extensions.serving.serde_v2 import (
    MlflowPyfuncHttpxSerializer,
    make_error_response,
)

LOGGER: Logger = get_logger()


@dataclass
class CustomEngineServingResponse:
    status: int
    data: dict


DIAGNOSTICS_REQUEST_KEY = "COMPUTE_DIAGNOSTICS"
ENABLE_DIAGNOSTICS_FLAG = "ENABLE_DIAGNOSTICS"
HEALTH_CHECK_KEY = "HEALTH_CHECK"

LOG_LEVEL: str = "LOG_LEVEL"
LOG_FILE_KEY: str = "LOG_FILE"
ARCHIVE_LOG_PATH_KEY: str = "ARCHIVE_LOG_PATH"


class CustomServingEnginePyfuncWrapper(mlflow.pyfunc.PythonModel):

    # todo support lora modules
    def __init__(self, *, engine: Type[EngineProcess], engine_config: EngineConfig):
        self._engine_klass: Type[EngineProcess] = engine
        self._engine_config: EngineConfig = engine_config
        self._engine: Optional[EngineProcess] = None
        # for convenience
        self._model_name = self._engine_config.model
        self._artifacts = None

    @property
    def artifacts(self):
        if self._artifacts is None:
            raise ValueError("Artifacts not configured, run model.setup()")
        return self._artifacts

    def _request_model(self, req: Request):
        # deserializer takes care of identifying whether oai client is needed or standard server client
        # so directly send to the root of the server
        try:
            response = self._engine.server_http_client.send(req)
        except Exception as e:
            LOGGER.error(
                f"Error in request: {req}, server_status: {self._engine.health_check_status()},"
                f" error: {str(e)}"
            )
            return MlflowPyfuncHttpxSerializer.serialize_response(
                make_error_response(
                    original_request=req,
                    error_message=str(e),
                    error_type=str(type(e)),
                    error_details={
                        "server_status": self._engine.health_check_status(),
                        "nvidia-smi": get_compute_details(cmd_key="nvidia-smi"),
                        "active-procs": get_compute_details(cmd_key="active-procs"),
                        "command": self._engine_config.to_run_command(),
                    },
                )
            )
        return MlflowPyfuncHttpxSerializer.serialize_response(response)

    @staticmethod
    def iter_mlflow_predictions(response: Response) -> Iterator[ResponseMessageV1]:
        mlflow_response = response.json()
        predictions = mlflow_response.get("predictions", [])
        for prediction in predictions:
            yield ResponseMessageV1.deserialize(prediction)

    def load_context(self, context: PythonModelContext):
        default_home_path = "~/.mlflow-extensions/logs/serving.log"
        resolved_path = os.path.expanduser(default_home_path)
        log_config: LogConfig = LogConfig(
            filename=os.environ.get(LOG_FILE_KEY, str(resolved_path)),
            archive_path=os.environ.get(ARCHIVE_LOG_PATH_KEY),
        )
        initialize_logging(log_config)

        if self._engine is None:
            self._engine = self._engine_klass(config=self._engine_config)

        self._engine.start_proc(context)

    def predict(
        self, context, model_input: List[List[str]], params=None
    ) -> List[List[str]]:
        if not isinstance(model_input, (list, dict, np.ndarray, pd.DataFrame)):
            raise ValueError(
                f"model_input must be a list, dict, numpy array, or dataframe but received: {type(model_input)}"
            )
        if isinstance(model_input, dict):
            model_input = model_input.values()
        if isinstance(model_input, pd.DataFrame):
            # check if its one column
            if len(model_input.columns) == 1:
                model_input = model_input[model_input.columns[0]].values
            else:
                raise ValueError(
                    f"Dataframe must have only one column, but received {len(model_input.columns)} columns"
                )

        responses = []
        for req in model_input:
            req: str
            if (
                req.startswith(DIAGNOSTICS_REQUEST_KEY)
                and os.environ.get(ENABLE_DIAGNOSTICS_FLAG, "false") == "true"
            ):
                parts = req.split(":")
                if len(parts) > 1:
                    compute_details = get_compute_details(parts[1])
                else:
                    compute_details = get_compute_details(cmd_key="all")
                compute_details.update(
                    {
                        "command": self._engine_config.to_run_command(context),
                    }
                )
                responses.append(json.dumps(compute_details))
            elif (
                req.startswith(DIAGNOSTICS_REQUEST_KEY)
                and os.environ.get(ENABLE_DIAGNOSTICS_FLAG, "false") == "false"
            ):
                responses.append(
                    f"Diagnostics are disabled please set environment variable on your deployment "
                    f"ENABLE_DIAGNOSTICS=true to enable them."
                )
            elif req.startswith(HEALTH_CHECK_KEY):
                responses.append(json.dumps(self._engine.health_check_status()))
            else:
                responses.append(
                    self._request_model(
                        MlflowPyfuncHttpxSerializer.deserialize_request(
                            req,
                            openai_base_url=self._engine.oai_http_client.base_url,
                            server_base_url=self._engine.server_http_client.base_url,
                        )
                    )
                )
        return responses

    def _setup_artifacts(self, local_dir: str = "/root/models"):
        self._artifacts = self._engine_config.setup_artifacts(local_dir)
        return self._artifacts

    def get_pip_reqs(self, **kwargs):
        return self._engine_config.default_pip_reqs(**kwargs)

    def setup(self, *, local_dir=None):
        if local_dir is None:
            home_directory = Path.home() / "models"
            home_directory.mkdir(parents=True, exist_ok=True)
        else:
            home_directory = local_dir
        LOGGER.info(f"Setting up artifacts in {home_directory}")
        self._setup_artifacts(str(home_directory))
        LOGGER.info(f"Command to be run: {self._engine_config.to_run_command()}")



from mlflow_extensions.serving.compat import inject_mlflow_openai_compat_client

MODULE = "openai"
INSTALL = "openai"

try:
    from openai import AsyncOpenAI, OpenAI
except ImportError as e:
    print(
        f"Error importing {MODULE} module please run "
        f"pip install {INSTALL} or upgrade the sdk by running pip install {INSTALL} --upgrade"
    )

OpenAI = inject_mlflow_openai_compat_client(use_sync=True)(OpenAI)
AsyncOpenAI = inject_mlflow_openai_compat_client(use_async=True)(AsyncOpenAI)



import json
import os
import queue
import signal
import site
import subprocess
import sys
import time
import typing
from threading import Thread
from typing import Any, Dict, List, Optional

import httpx

from mlflow_extensions.serving.serde_v2 import MlflowPyfuncHttpxSerializer
from mlflow_extensions.testing.helper import (
    is_process_active,
    kill_processes_containing,
)

if typing.TYPE_CHECKING:
    from openai import OpenAI

from mlflow_extensions.log import Logger, get_logger

LOGGER: Logger = get_logger()


class FixedSizeLogQueue(queue.Queue):
    def __init__(self, max_size: int = 10000):
        # support at most 10000 log-lines
        super().__init__(maxsize=max_size)

    def put(self, item, block=True, timeout=None):
        if self.full():
            self.get_nowait()
        super().put(item, block, timeout)


def ensure_proper_site_packages_for_local(env: dict):
    python_exec_path = sys.executable
    bin_dir = os.path.dirname(python_exec_path)
    site_packages_path = site.getsitepackages()[0]
    LOGGER.info(f"Setting PYTHONPATH to include {site_packages_path}")
    LOGGER.info(f"Setting PATH to include {bin_dir}")
    if env.get("PYTHONPATH", "").startswith(site_packages_path) is False:
        env["PYTHONPATH"] = f"{site_packages_path}:{env.get('PYTHONPATH', '')}"
    if env.get("PATH", "").startswith(bin_dir) is False:
        env["PATH"] = f"{bin_dir}:{env.get('PATH', '')}"


class LocalTestServer:

    def __init__(
        self,
        *,
        model_uri: str,
        registry_host: str,
        registry_token: str,
        test_serving_host: str = "0.0.0.0",
        test_serving_port: int = 5000,
        registry_is_uc: bool = False,
        additional_serving_flags: Optional[List[str]] = None,
        use_local_env: bool = False,
        wait_server_start_interval_seconds: int = 5,
    ):
        self._wait_server_start_interval_seconds = wait_server_start_interval_seconds
        self._use_local_env = use_local_env
        self._model_uri = model_uri
        self._databricks_registry_host = registry_host
        self._databricks_registry_token = registry_token
        self._test_serving_host = test_serving_host
        self._test_serving_port = test_serving_port
        self._registry_is_uc = registry_is_uc
        self._additional_serving_flags = additional_serving_flags or []
        if (
            self._use_local_env is True
            and "--env-manager" not in self._additional_serving_flags
        ):
            self._additional_serving_flags.append("--env-manager")
            self._additional_serving_flags.append("local")

        self._server_process: Optional[subprocess.Popen] = None
        self._http_client = httpx.Client(
            base_url=f"http://{self._test_serving_host}:{self._test_serving_port}"
        )
        self._log_queue = FixedSizeLogQueue(max_size=10000)

    def start(self):
        kill_processes_containing("mlflow models serve")
        command_args = [
            "mlflow",
            "models",
            "serve",
            "-m",
            self._model_uri,
            "-p",
            str(self._test_serving_port),
            "-w",
            "4",
            *self._additional_serving_flags,
        ]
        LOGGER.info(f"Starting server with command: {' '.join(command_args)}")
        # spawn in new process group
        current_env = os.environ.copy()
        if self._use_local_env is True:
            ensure_proper_site_packages_for_local(current_env)
        current_env["DATABRICKS_HOST"] = self._databricks_registry_host
        current_env["DATABRICKS_TOKEN"] = self._databricks_registry_token
        self._server_process = subprocess.Popen(
            command_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=os.setsid,
            env=current_env,
        )
        Thread(
            target=self._enqueue_output,
            args=(self._server_process.stdout, self._log_queue),
        ).start()
        Thread(
            target=self._enqueue_output,
            args=(self._server_process.stderr, self._log_queue),
        ).start()

    def __enter__(self):
        self.start()
        LOGGER.info(
            f"Started server on {self._test_serving_host}:{self._test_serving_port} for model {self._model_uri}"
        )
        LOGGER.info("Waiting for server to be healthy...")
        self.wait_and_assert_healthy(self._wait_server_start_interval_seconds)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            LOGGER.info(f"Exception: {exc_type} {exc_val}")
        self.stop()
        LOGGER.info("Stopped server")

    def _enqueue_output(self, pipe, q: queue.Queue):
        for line in iter(pipe.readline, b""):
            q.put(line)
        pipe.close()

    def _flush_current_logs(self, max_wait_time_seconds: int = 10):
        time_start = time.time()

        while True:
            # ensure that we get to other parts of the code
            if time.time() - time_start > max_wait_time_seconds:
                break
            try:
                line = self._log_queue.get_nowait()
                data = line.decode().strip()
                if len(data) > 0:
                    print(data)
            except queue.Empty:
                break

    def wait_and_assert_healthy(self, wait_interval_seconds: int = 5):
        assert self._server_process is not None, "Server process has not been started."
        while True:
            self._flush_current_logs()

            if is_process_active(self._server_process.pid) is False:
                stdout, stderr = self._server_process.communicate(timeout=10)
                LOGGER.info(f"STDOUT: {stdout.decode()}")
                LOGGER.info(f"STDERR: {stderr.decode()}")
                LOGGER.info(f"Exit Code: {self._server_process.returncode}")
                raise ValueError("Server process has terminated unexpectedly.")

            try:
                resp = self._http_client.get("/health")
                if resp.status_code == 200:
                    LOGGER.info("Success")
                    break
            except Exception as e:
                LOGGER.warning(
                    f"[HEALTH_CHECK] endpoint not yet available; health check error {str(e)}"
                )

            time.sleep(wait_interval_seconds)

            # assert (
            #     self._server_process is not None and self._server_process.pid is not None
            #     and is_process_active(self._server_process.pid) is True
            # ), "Server process has terminated unexpectedly."

    def query(self, *, payload: Dict[str, Any], timeout: int = 30):
        return self._http_client.post("/invocations", json=payload, timeout=timeout)

    def query_custom_server(
        self,
        *,
        method: str,
        http_path: str,
        headers: dict = None,
        api_payload: Dict[str, Any] = None,
        timeout: int = 30,
        is_openai_compatible: bool = False,
    ):
        self._flush_current_logs()
        orig_request = httpx.Request(
            method=method,
            url=f"http://{self._test_serving_host}:{self._test_serving_port}{http_path}",
            headers=headers or {},
            content=json.dumps(api_payload) if api_payload else None,
        )
        response = self._http_client.post(
            "/invocations",
            json={
                "inputs": [
                    MlflowPyfuncHttpxSerializer.serialize_request(
                        request=orig_request,
                        url_path_to_request=http_path,
                        requires_openai_compat=is_openai_compatible,
                    )
                ]
            },
            timeout=timeout,
        )
        if response.status_code != 200:
            print(response.json())
        predictions = response.json()["predictions"]
        self._flush_current_logs()
        return MlflowPyfuncHttpxSerializer.deserialize_response(
            predictions[0], orig_request
        )

    @property
    def openai_client(self) -> "OpenAI":
        from mlflow_extensions.serving.compat.openai import OpenAI

        return OpenAI(
            base_url=f"http://{self._test_serving_host}:{self._test_serving_port}/invocations",
            api_key="foobar",
        )

    def stop(self):
        self._flush_current_logs()
        os.killpg(os.getpgid(self._server_process.pid), signal.SIGTERM)
        time.sleep(5)
        if self._server_process.poll() is None:
            os.killpg(os.getpgid(self._server_process.pid), signal.SIGKILL)
        try:
            stdout, stderr = self._server_process.communicate(timeout=10)
        except subprocess.TimeoutExpired:
            stdout, stderr = self._server_process.communicate()
            print("Timed out")

        print("STDOUT:", stdout.decode())
        print("STDERR:", stderr.decode())
        print("Exit Code:", self._server_process.returncode)



import typing

from mlflow_extensions.testing.helper import Modality, inject_openai_client, run_if

if typing.TYPE_CHECKING is True:
    from openai import OpenAi

    from mlflow_extensions.testing.runner import ModelContextRunner


@run_if(modality=Modality.TEXT.value)
@inject_openai_client
def query_text(
    *,
    ctx: "ModelContextRunner",
    client: "OpenAi",
    model: str,
    modality_type: str = None,
    host: str = "0.0.0.0",  # noqa
    port: int = 9989,  # noqa
    repeat_n: int = 5,
):
    count = 0
    # repeat a few times
    while count < repeat_n:
        from typing import List, Literal

        from pydantic import BaseModel

        class ExtractedBody(BaseModel):
            product: str
            languages: Literal["python", "sql", "scala"]
            keywords: List[str]
            strategies: List[str]

        try:
            response = client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": """The Databricks Lakehouse Platform for Dummies is your guide to simplifying 
            your data storage. The lakehouse platform has SQL and performance 
            capabilities - indexing, caching and MPP processing - to make 
            BI work rapidly on data lakes. It also provides direct file access 
            and direct native support for Python, data science and 
            AI frameworks without the need to force data through an 
            SQL-based data warehouse. Find out how the lakehouse platform 
            creates an opportunity for you to accelerate your data strategy.""",
                    }
                ],
                extra_body={"guided_json": ExtractedBody.schema()},
            )
            ctx.add_success(result=response.choices[0].message.content)
        except Exception as e:
            print(e)
            ctx.add_error(error_msg=str(e))
        finally:
            count += 1



import json
from base64 import b64decode, b64encode
from io import BytesIO
from typing import Any, Dict

import httpx
from httpx import URL, Request


class RequestSerdeKeys:
    METHOD = "method"
    URL_PATH = "url_path"
    HEADERS = "headers"
    BODY = "body"
    BODY_TYPE = "body_type"
    REQUIRES_OPENAI_COMPAT = "requires_openai_compat"


class ResponseSerdeKeys:
    STATUS_CODE = "status_code"
    HEADERS = "headers"
    CONTENT = "content"
    CONTENT_TYPE = "content_type"


class SerdeContentTypes:
    MULTIPART = "multipart"
    TEXT = "text"
    STREAM = "stream"
    NONE = "none"


class MlflowPyfuncHttpxSerializer:
    @staticmethod
    def serialize_request(
        request: Request, url_path_to_request: str, requires_openai_compat: bool = True
    ):
        headers = dict(request.headers)

        if headers["content-type"].startswith("multipart/form-data"):
            body_type = SerdeContentTypes.MULTIPART
            body = request.read().decode("utf-8")
        elif isinstance(request.content, str):
            body = request.content
            body_type = SerdeContentTypes.TEXT
        elif isinstance(request.content, bytes):
            if request.headers.get("content-type", "").startswith(
                "application/json"
            ) or request.headers.get("content-type", "").startswith("text/plain"):
                body = request.content.decode("utf-8")
                body_type = SerdeContentTypes.TEXT
            else:
                body = b64encode(request.content).decode("utf-8")
                body_type = SerdeContentTypes.STREAM
        elif request.content is None:
            body = None
            body_type = SerdeContentTypes.NONE
        else:
            body = request.read()
            body = b64encode(body).decode("utf-8")
            body_type = SerdeContentTypes.STREAM

        serialized = {
            RequestSerdeKeys.METHOD: request.method,
            RequestSerdeKeys.URL_PATH: url_path_to_request,
            RequestSerdeKeys.HEADERS: {
                "content-type": headers.get("content-type", "text/plain")
            },
            RequestSerdeKeys.BODY: body,
            RequestSerdeKeys.BODY_TYPE: body_type,
            RequestSerdeKeys.REQUIRES_OPENAI_COMPAT: requires_openai_compat,
        }
        return json.dumps(serialized)

    @staticmethod
    def deserialize_request(
        serialized_request: str, *, openai_base_url: URL, server_base_url: URL
    ):
        req_data = json.loads(serialized_request)

        if req_data[RequestSerdeKeys.BODY_TYPE] == SerdeContentTypes.STREAM:
            content = BytesIO(b64decode(req_data[RequestSerdeKeys.BODY]))
        elif req_data[RequestSerdeKeys.BODY_TYPE] == SerdeContentTypes.MULTIPART:
            content = BytesIO(req_data[RequestSerdeKeys.BODY].encode("utf-8"))
        elif req_data[RequestSerdeKeys.BODY_TYPE] == SerdeContentTypes.NONE:
            content = None
        else:
            content = req_data[RequestSerdeKeys.BODY]

        if req_data[RequestSerdeKeys.REQUIRES_OPENAI_COMPAT] is True:
            base_url = openai_base_url
        else:
            base_url = server_base_url
        full_url = base_url.join(req_data[RequestSerdeKeys.URL_PATH].lstrip("/"))

        return httpx.Request(
            method=req_data[RequestSerdeKeys.METHOD],
            url=full_url,
            headers=req_data[RequestSerdeKeys.HEADERS],
            content=content,
        )

    @staticmethod
    def serialize_response(response):
        headers = dict(response.headers)

        # Read and encode the entire content
        if headers.get("content-type", "").startswith(
            "application/json"
        ) or headers.get("content-type", "").startswith("text/plain"):
            content = response.read().decode("utf-8")
            content_type = SerdeContentTypes.TEXT
        else:
            content = response.read()
            content = b64encode(content).decode("utf-8")
            content_type = SerdeContentTypes.STREAM

        serialized = {
            ResponseSerdeKeys.STATUS_CODE: response.status_code,
            ResponseSerdeKeys.HEADERS: headers,
            ResponseSerdeKeys.CONTENT: content,
            ResponseSerdeKeys.CONTENT_TYPE: content_type,
        }
        return json.dumps(serialized)

    @staticmethod
    def deserialize_response(serialized_response, orig_request: Request = None):
        resp_data = json.loads(serialized_response)

        if resp_data[ResponseSerdeKeys.CONTENT_TYPE] == SerdeContentTypes.STREAM:
            content = b64decode(resp_data[ResponseSerdeKeys.CONTENT])
        else:
            content = resp_data[ResponseSerdeKeys.CONTENT]

        return httpx.Response(
            status_code=resp_data[ResponseSerdeKeys.STATUS_CODE],
            headers=resp_data[ResponseSerdeKeys.HEADERS],
            content=content,
            request=orig_request,
        )


def make_error_response(
    *,
    original_request: Request,
    error_message: str,
    error_type: str,
    error_details: Dict[str, Any],
    status_code: int = 500
):
    return httpx.Response(
        status_code=status_code,
        headers={"content-type": "application/json"},
        content=json.dumps(
            {
                "error": error_message,
                "error_type": error_type,
                "error_details": error_details,
                "original_request": MlflowPyfuncHttpxSerializer.serialize_request(
                    original_request, url_path_to_request=original_request.url.path
                ),
            }
        ),
    )



from dataclasses import dataclass, field

from mlflow_extensions.databricks.prebuilt.audio import AudioLM
from mlflow_extensions.databricks.prebuilt.text import TextLM
from mlflow_extensions.databricks.prebuilt.vision import VisionLM


@dataclass(frozen=True)
class Prebuilt:
    text: TextLM = field(default_factory=lambda: TextLM())
    vision: VisionLM = field(default_factory=lambda: VisionLM())
    audio: AudioLM = field(default_factory=lambda: AudioLM())


prebuilt = Prebuilt()



import functools
import inspect
import logging
import os
import socket
import sys
from dataclasses import dataclass, field
from enum import IntEnum
from types import FrameType
from typing import Any, Callable, Dict, List, Optional, Union

import structlog
import structlog.stdlib
from pythonjsonlogger import jsonlogger
from structlog.stdlib import BoundLogger
from structlog.types import EventDict, WrappedLogger

from mlflow_extensions.log.handlers import rotating_volume_handler
from mlflow_extensions.version import get_mlflow_extensions_version

Logger = BoundLogger

DEFAULT_MAX_BYTES: int = 10 * 1024 * 1024
DEFAULT_BACKUP_COUNT: int = 5


class LogLevel(IntEnum):
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    @classmethod
    def from_int(cls, value: int):
        try:
            return cls(value)
        except ValueError:
            raise ValueError(f"No LogLevel with integer value {value}")

    @classmethod
    def from_string(cls, name: str):
        try:
            return cls[name.upper()]
        except KeyError:
            raise ValueError(f"No LogLevel with name {name}")

    @classmethod
    def to_level(cls, level: Union["LogLevel", str, int]) -> "LogLevel":
        if isinstance(level, LogLevel):
            return level
        elif isinstance(level, str):
            return cls.from_string(level)
        elif isinstance(level, int):
            return cls.from_int(level)
        else:
            raise ValueError(f"Invalid log level: {level}")


@dataclass
class LogConfig:
    filename: Optional[str] = None
    level: Union[LogLevel, int, str] = LogLevel.INFO
    archive_path: Optional[str] = None
    databricks_host: Optional[str] = None
    databricks_token: Optional[str] = None
    when: str = "m"
    interval: int = 5
    max_bytes: int = DEFAULT_MAX_BYTES
    backup_count: int = DEFAULT_BACKUP_COUNT
    encoding: Optional[str] = None
    delay: bool = False
    utc: bool = False
    at_time = None
    errors = None
    additional_vars: Dict[str, Any] = field(default_factory=dict)


def get_logger(name: Optional[str] = None) -> Logger:
    """
    Returns a logger with the specified name.

    Parameters:
        name (Optional[str]): The name of the logger. If not provided, the name will be determined based on the caller's module and class (if applicable).

    Returns:
        Logger: The logger object.

    Example:
        logger: Logger = get_logger()
    """

    if name is None:
        frame: Optional[FrameType] = inspect.currentframe()
        caller_frame: Optional[FrameType] = frame.f_back
        module: Dict[str, Any] = caller_frame.f_globals["__name__"]

        # Check if it's a method inside a class and get the class name
        if "self" in caller_frame.f_locals:
            class_name: str = caller_frame.f_locals["self"].__class__.__name__
            name = f"{module}.{class_name}"
        else:
            name = module

    logger: Logger = structlog.stdlib.get_logger(name)

    return logger


def initialize_logging(config: LogConfig) -> None:

    stdout_handler: logging.Handler = logging.StreamHandler(sys.stdout)
    handlers: List[logging.Handler] = [stdout_handler]

    if config.filename is not None:
        volume_handler: logging.Handler = rotating_volume_handler(
            filename=config.filename,
            archive_path=config.archive_path,
            max_bytes=config.max_bytes,
            backup_count=config.backup_count,
            databricks_host=config.databricks_host,
            databricks_token=config.databricks_token,
            when=config.when,
            interval=config.interval,
        )
        handlers.append(volume_handler)

    level: LogLevel = LogLevel.to_level(config.level)
    json_formatter: jsonlogger.JsonFormatter = jsonlogger.JsonFormatter()

    for handler in handlers:
        handler.setFormatter(json_formatter)
        handler.setLevel(level)

    logging.shutdown()
    logging.basicConfig(level=level, format="%(message)s", handlers=handlers)

    def filter_by_level(logger: Logger, name: str, event_dict: EventDict) -> EventDict:
        event_log_level: LogLevel = LogLevel.from_string(name)
        if event_log_level >= level:
            return event_dict
        else:
            raise structlog.DropEvent

    def add_additional_vars(
        logger: Logger, method_name: str, event_dict: EventDict
    ) -> EventDict:
        for key, value in config.additional_vars.items():
            event_dict[key] = value
        return event_dict

    structlog.configure(
        processors=[
            filter_by_level,
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            add_additional_vars,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.ExceptionRenderer(),
            structlog.processors.UnicodeDecoder(),
            # structlog.processors.CallsiteParameterAdder(
            #     {
            #         structlog.processors.CallsiteParameter.FILENAME,
            #         structlog.processors.CallsiteParameter.FUNC_NAME,
            #         structlog.processors.CallsiteParameter.LINENO,
            #     }
            # ),
            # structlog.processors.JSONRenderer(),
            structlog.stdlib.render_to_log_kwargs,
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def log_around(_func=None, *, logger: Union[Logger, logging.Logger] = None):
    def decorator_log(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal logger
            if logger is None:
                logger = get_logger(func.__name__)
            args_repr: List[str] = [repr(a) for a in args]
            kwargs_repr: List[str] = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature: str = ", ".join(args_repr + kwargs_repr)
            logger.debug(f"entry: {func.__name__}", signature=signature)
            try:
                result = func(*args, **kwargs)
                logger.debug(f"exit: {func.__name__}", result=result)
                return result
            except Exception as e:
                logger.exception(
                    f"Exception raised in {func.__name__}. exception: {str(e)}"
                )
                raise e

        return wrapper

    if _func is None:
        return decorator_log
    else:
        return decorator_log(_func)



from dataclasses import dataclass, field
from typing import Optional

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig, ServingConfig
from mlflow_extensions.serving.engines import VLLMEngineConfig, VLLMEngineProcess

_ENGINE = VLLMEngineProcess
_ENGINE_CONFIG = VLLMEngineConfig


def phi_3_5_vision_instruct(
    ctx_name: str,
    context_length: Optional[int],
    min_memory_gb: int,
    max_num_images: int = 1,
) -> EzDeployConfig:
    model_cfg = {
        "max_model_len": context_length,
    }
    return EzDeployConfig(
        name=f"phi_3_5_vision_instruct_{ctx_name}",
        engine_proc=_ENGINE,
        engine_config=_ENGINE_CONFIG(
            model="microsoft/Phi-3.5-vision-instruct",
            trust_remote_code=True,
            guided_decoding_backend="outlines",
            max_num_images=max_num_images,
            vllm_command_flags={
                "--gpu-memory-utilization": 0.98,
                "--enforce-eager": None,
            },
            **model_cfg,
        ),
        serving_config=ServingConfig(
            # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
            minimum_memory_in_gb=min_memory_gb,
        ),
    )


# https://huggingface.co/microsoft/Phi-3.5-vision-instruct
PHI_3_5_VISION_INSTRUCT_4K_CONFIG = phi_3_5_vision_instruct(
    "4k", 4096, 12, max_num_images=1
)

# https://huggingface.co/microsoft/Phi-3.5-vision-instruct
PHI_3_5_VISION_INSTRUCT_8K_CONFIG = phi_3_5_vision_instruct(
    "8k", 8192, 15, max_num_images=2
)

# https://huggingface.co/microsoft/Phi-3.5-vision-instruct
PHI_3_5_VISION_INSTRUCT_12K_CONFIG = phi_3_5_vision_instruct(
    "12k", 12000, 20, max_num_images=4
)

# https://huggingface.co/microsoft/Phi-3.5-vision-instruct
PHI_3_5_VISION_INSTRUCT_32K_CONFIG = phi_3_5_vision_instruct(
    "32k", 32000, 30, max_num_images=4
)

# https://huggingface.co/microsoft/Phi-3.5-vision-instruct
PHI_3_5_VISION_INSTRUCT_64K_CONFIG = phi_3_5_vision_instruct(
    "64k", 64000, 50, max_num_images=8
)

# https://huggingface.co/microsoft/Phi-3.5-vision-instruct
PHI_3_5_VISION_INSTRUCT_128K_CONFIG = phi_3_5_vision_instruct(
    "128k", None, 110, max_num_images=8
)

QWEN2_VL_2B_INSTRUCT = EzDeployConfig(
    name="qwen2_vl_2b_instruct",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="Qwen/Qwen2-VL-2B-Instruct",
        guided_decoding_backend="outlines",
        vllm_command_flags={
            "--gpu-memory-utilization": 0.98,
            "--distributed-executor-backend": "ray",
        },
        max_num_images=5,
        # TODO FIX THIS ON NEW RELEASE OF TRANSFORMERS 0.45.0 otherwise you will get qwen2_vl not found
        # https://github.com/huggingface/transformers/issues/33401
        # https://github.com/QwenLM/Qwen2-VL?tab=readme-ov-file#quickstart
        library_overrides={
            "transformers": "git+https://github.com/huggingface/transformers@21fac7abba2a37fae86106f87fcf9974fd1e3830",
            "accelerate": "accelerate==0.31.0",
            "vllm": "vllm==0.6.1",
        },
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=20,
    ),
)

QWEN2_VL_7B_INSTRUCT = EzDeployConfig(
    name="qwen2_vl_7b_instruct",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="Qwen/Qwen2-VL-7B-Instruct",
        guided_decoding_backend="outlines",
        vllm_command_flags={
            "--gpu-memory-utilization": 0.98,
            "--distributed-executor-backend": "ray",
        },
        max_num_images=5,
        # TODO FIX THIS ON NEW RELEASE OF TRANSFORMERS 0.45.0 otherwise you will get qwen2_vl not found
        # https://github.com/huggingface/transformers/issues/33401
        # https://github.com/QwenLM/Qwen2-VL?tab=readme-ov-file#quickstart
        library_overrides={
            "transformers": "git+https://github.com/huggingface/transformers@21fac7abba2a37fae86106f87fcf9974fd1e3830",
            "accelerate": "accelerate==0.31.0",
            "vllm": "vllm==0.6.1",
        },
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=70,
    ),
)


@dataclass(frozen=True)
class VllmVision:
    PHI_3_5_VISION_INSTRUCT_4K: EzDeployConfig = field(
        default_factory=lambda: PHI_3_5_VISION_INSTRUCT_4K_CONFIG
    )
    PHI_3_5_VISION_INSTRUCT_8K: EzDeployConfig = field(
        default_factory=lambda: PHI_3_5_VISION_INSTRUCT_8K_CONFIG
    )
    PHI_3_5_VISION_INSTRUCT_12K: EzDeployConfig = field(
        default_factory=lambda: PHI_3_5_VISION_INSTRUCT_12K_CONFIG
    )
    PHI_3_5_VISION_INSTRUCT_32K: EzDeployConfig = field(
        default_factory=lambda: PHI_3_5_VISION_INSTRUCT_32K_CONFIG
    )
    PHI_3_5_VISION_INSTRUCT_64K: EzDeployConfig = field(
        default_factory=lambda: PHI_3_5_VISION_INSTRUCT_64K_CONFIG
    )
    PHI_3_5_VISION_INSTRUCT_128K: EzDeployConfig = field(
        default_factory=lambda: PHI_3_5_VISION_INSTRUCT_128K_CONFIG
    )
    QWEN2_VL_2B_INSTRUCT: EzDeployConfig = field(
        default_factory=lambda: QWEN2_VL_2B_INSTRUCT
    )
    QWEN2_VL_7B_INSTRUCT: EzDeployConfig = field(
        default_factory=lambda: QWEN2_VL_7B_INSTRUCT
    )



import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union

from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.log import Logger, get_logger
from mlflow_extensions.serving.engines import gpu_utils
from mlflow_extensions.serving.engines.base import Command, EngineConfig, EngineProcess
from mlflow_extensions.serving.engines.huggingface_utils import (
    ensure_chat_template,
    snapshot_download_local,
)
from mlflow_extensions.testing.helper import kill_processes_containing

LOGGER: Logger = get_logger()


@dataclass(frozen=True, kw_only=True)
class VLLMEngineConfig(EngineConfig):
    entrypoint_module: str = field(default="vllm.entrypoints.openai.api_server")
    enable_experimental_chunked_prefill: bool = field(default=False)
    max_num_batched_tokens: int = field(
        default=None
    )  # 512 is based on A100 ITL for llama model
    enable_prefix_caching: bool = field(default=False)
    vllm_command_flags: Dict[str, Optional[str]] = field(default_factory=dict)
    trust_remote_code: bool = field(default=False)
    max_model_len: Optional[int] = field(default=None)
    served_model_alias: Optional[str] = field(default=None)
    guided_decoding_backend: Optional[str] = field(default="outlines")
    tokenizer: Optional[str] = field(default=None)  # --tokenizer

    max_num_images: Optional[int] = field(default=None)
    max_num_videos: Optional[int] = field(default=None)
    max_num_audios: Optional[int] = field(default=None)

    # general keys
    model_artifact_key: str = field(default="model")
    verify_chat_template: bool = field(default=True)
    tokenizer_artifact_key: str = field(default="tokenizer")
    tokenizer_config_file: str = field(default="tokenizer_config.json")
    chat_template_key: str = field(default="chat_template")
    tokenizer_mode: Optional[str] = field(default=None)

    def _to_vllm_command(self, context: PythonModelContext = None) -> List[str]:
        local_model_path = None
        tokenizer_path = None
        if context is not None:
            local_model_path = context.artifacts.get(self.model_artifact_key)
            tokenizer_path = context.artifacts.get(self.tokenizer_artifact_key)
        flags = []
        skip_flags = [
            "--enable-chunked-prefill",
            "--model",
            "--tokenizer",
            "--max-num-batched-tokens",
            "--enable-prefix-caching",
            "--max-model-len",
            "--trust-remote-code",
            "--served-model-name",
            "--guided-decoding-backend",
            "--limit-mm-per-prompt",
            "--tokenizer-mode",
        ]

        # add tensor parallel size flag if we have GPUs
        gpu_count = gpu_utils.get_gpu_count()
        if gpu_count >= 1:
            flags.append("--tensor-parallel-size")
            flags.append(str(gpu_count))

        for k, v in self.vllm_command_flags.items():
            if k in skip_flags:
                LOGGER.info(f"Skipping flag {k} use the built in argument")
                continue
            flags.append(k)
            if v is not None:
                flags.append(v)
        if self.enable_experimental_chunked_prefill is True:
            flags.append("--enable-chunked-prefill")
            if self.max_num_batched_tokens is None:
                flags.append("--max-num-batched-tokens")
                flags.append("512")
        if self.max_num_batched_tokens is not None:
            flags.append("--max-num-batched-tokens")
            flags.append(str(self.max_num_batched_tokens))
        if self.enable_prefix_caching is True:
            flags.append("--enable-prefix-caching")
        if self.trust_remote_code is True:
            flags.append("--trust-remote-code")
        if self.max_model_len is not None:
            flags.append("--max-model-len")
            flags.append(str(self.max_model_len))
        if self.guided_decoding_backend is not None:
            flags.append("--guided-decoding-backend")
            if self.guided_decoding_backend not in ["outlines", "lm-format-enforcer"]:
                raise ValueError(
                    f"Invalid guided decoding backend {self.guided_decoding_backend}"
                )
            flags.append(self.guided_decoding_backend)
        if self.tokenizer is not None:
            flags.append("--tokenizer")
            flags.append(tokenizer_path or self.tokenizer)
        if (
            self.max_num_images is not None
            or self.max_num_videos is not None
            or self.max_num_audios is not None
        ):
            flags.append("--limit-mm-per-prompt")
            values = []
            if self.max_num_images is not None:
                values.append(f"image={self.max_num_images}")
            if self.max_num_videos is not None:
                values.append(f"video={self.max_num_videos}")
            if self.max_num_audios is not None:
                values.append(f"audio={self.max_num_audios}")
            flags.append(",".join(values))
        if self.tokenizer_mode is not None:
            flags.append("--tokenizer-mode")
            flags.append(self.tokenizer_mode)

        return [
            sys.executable,
            "-m",
            self.entrypoint_module,
            "--host",
            self.host,
            "--port",
            str(self.port),
            "--model",
            self.model if local_model_path is None else local_model_path,
            "--served-model-name",
            self.model if self.served_model_alias is None else self.served_model_alias,
            *flags,
        ]

    def _to_run_command(
        self, context: PythonModelContext = None
    ) -> Union[List[str], Command]:
        return self._to_vllm_command(context=context)

    def engine_pip_reqs(
        self,
        *,
        vllm_version: str = "0.6.0",
        lm_format_enforcer_version: str = "0.10.6",
        outlines_version: str = "0.0.46",
    ) -> Dict[str, str]:
        default_installs = {
            "vllm": f"vllm=={vllm_version}",
        }
        if self.guided_decoding_backend == "lm-format-enforcer":
            default_installs["lm-format-enforcer"] = (
                f"lm-format-enforcer=={lm_format_enforcer_version}"
            )
        if self.guided_decoding_backend == "outlines":
            default_installs["outlines"] = f"outlines=={outlines_version}"
        return default_installs

    def _setup_snapshot(self, local_dir: str = "/root/models"):
        return snapshot_download_local(repo_id=self.model, local_dir=local_dir)

    def _setup_artifacts(self, local_dir: str = "/root/models"):
        local_path = self._setup_snapshot(local_dir)
        # tokenizer path
        artifacts = {self.model_artifact_key: local_path}
        if self.tokenizer is not None and self.model != self.tokenizer:
            tokenizer_local_path = snapshot_download_local(
                repo_id=self.tokenizer, local_dir=local_dir, tokenizer_only=True
            )
            artifacts[self.tokenizer_artifact_key] = tokenizer_local_path
        return artifacts

    def _verify_chat_template(self, artifacts: Dict[str, str]):
        if self.tokenizer is None:
            model_dir_path = Path(artifacts[self.model_artifact_key])
        else:
            model_dir_path = Path(artifacts[self.tokenizer_artifact_key])
        if self.tokenizer_mode is not None and self.tokenizer_mode in [
            "auto",
            "slow",
        ]:
            tokenizer_config_file = model_dir_path / self.tokenizer_config_file
            ensure_chat_template(
                tokenizer_file=str(tokenizer_config_file),
                chat_template_key=self.chat_template_key,
            )

    def setup_artifacts(self, local_dir: str = "/root/models"):
        artifacts = self._setup_artifacts(local_dir)
        if self.verify_chat_template is True:
            self._verify_chat_template(artifacts)
        return artifacts

    def supported_model_architectures(self) -> List[str]:
        return [
            "AquilaForCausalLM",
            "ArcticForCausalLM",
            "BaiChuanForCausalLM",
            "BloomForCausalLM",
            "ChatGLMModel",
            "CohereForCausalLM",
            "DbrxForCausalLM",
            "DeciLMForCausalLM",
            "ExaoneForCausalLM",
            "FalconForCausalLM",
            "GemmaForCausalLM",
            "Gemma2ForCausalLM",
            "GPT2LMHeadModel",
            "GPTBigCodeForCausalLM",
            "GPTJForCausalLM",
            "GPTNeoXForCausalLM",
            "InternLMForCausalLM",
            "InternLM2ForCausalLM",
            "JAISLMHeadModel",
            "JambaForCausalLM",
            "LlamaForCausalLM",
            "MiniCPMForCausalLM",
            "MistralForCausalLM",
            "MixtralForCausalLM",
            "MPTForCausalLM",
            "NemotronForCausalLM",
            "OLMoForCausalLM",
            "OPTForCausalLM",
            "OrionForCausalLM",
            "PhiForCausalLM",
            "Phi3ForCausalLM",
            "Phi3SmallForCausalLM",
            "PhiMoEForCausalLM",
            "PersimmonForCausalLM",
            "QWenLMHeadModel",
            "Qwen2ForCausalLM",
            "Qwen2MoeForCausalLM",
            "StableLmForCausalLM",
            "Starcoder2ForCausalLM",
            "XverseForCausalLM",
            "Blip2ForConditionalGeneration",
            "ChameleonForConditionalGeneration",
            "FuyuForCausalLM",
            "InternVLChatModel",
            "LlavaForConditionalGeneration",
            "LlavaNextForConditionalGeneration",
            "PaliGemmaForConditionalGeneration",
            "Phi3VForCausalLM",
            "MiniCPMV",
            "UltravoxModel",
        ]


class VLLMEngineProcess(EngineProcess):

    @property
    def engine_name(self) -> str:
        return "vllm-engine"

    def health_check(self) -> bool:
        try:
            resp = self.server_http_client.get("/health")
            return resp.status_code == 200
        except Exception as e:
            LOGGER.error(
                f"Health check failed with error {e}; server may not be up yet or crashed;"
            )
            return False

    def cleanup(self) -> None:
        try:
            import ray

            ray.shutdown()
        except Exception:
            pass

        try:
            kill_processes_containing("vllm")
            kill_processes_containing("ray")
            kill_processes_containing("from multiprocessing")
        except Exception:
            pass



# Databricks notebook source
# MAGIC %pip install mlflow-extensions==0.12.0
# MAGIC %pip install -U mlflow
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeploy
from mlflow_extensions.databricks.prebuilt import prebuilt
import os

os.environ["HF_TOKEN"] = dbutils.secrets.get(
    scope="sri-mlflow-extensions", key="hf-token"
)

deployer = EzDeploy(
  config=prebuilt.vision.vllm.QWEN2_VL_7B_INSTRUCT,
  registered_model_name="main.default.qwen2_vl_7b_instruct"
)

deployer.download()

deployer.register()

endpoint_name = "sri_qwen2_vl_7b_instruct"

deployer.deploy(endpoint_name, scale_to_zero=False)

# COMMAND ----------

endpoint_name = "sri_qwen2_vl_7b_instruct"

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow.utils.databricks_utils import get_databricks_host_creds
from mlflow_extensions.databricks.prebuilt import prebuilt
from pydantic import BaseModel
import typing as t

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"
token = get_databricks_host_creds().token

client = OpenAI(
  base_url=endpoint_name,
  api_key=token
)


class Comparison(BaseModel):
  image_1_details: str
  image_2_details: str
  image_1_colors: t.List[str]
  image_2_colors: t.List[str]
  image_1_has_human: bool
  image_2_has_human: bool
  image_1_human_gender: t.Literal["male", "female", "no human"]
  image_2_human_gender: t.Literal["male", "female", "no human"]

prompt = f"Compare the two images and use this schema for reference and no markdown just valid json: {Comparison.schema()}"



url_1 = "https://richmedia.ca-richimage.com/ImageDelivery/imageService?profileId=12026540&id=1859027&recipeId=728"
url_2 = "https://m.media-amazon.com/images/I/81W3YQdu-tL._AC_SY550_.jpg"

response = client.chat.completions.create(
    model=prebuilt.vision.vllm.QWEN2_VL_7B_INSTRUCT.engine_config.model,
    messages=[
      {
        "role": "user",
        "content": [
            {"type" : "text", "text": prompt},
            {"type": "image_url", "image_url": {"url": url_1}},
            {"type" : "text", "text": "to this image. Answer in english."},
            {"type": "image_url", "image_url": {"url": url_2}}
        ]
      }
    ],
    max_tokens=8192,
)
print(response.choices[0].message.content)

# COMMAND ----------





# Databricks notebook source
# MAGIC %md 
# MAGIC
# MAGIC # KEEP IN MIND AYA does not have a apache 2.0 or MIT license it is cc-by-nc-4.0 
# MAGIC
# MAGIC # YOU MUST adhere to C4AI's Acceptable Use Policy
# MAGIC
# MAGIC Read more here: https://huggingface.co/CohereForAI/aya-23-35B

# COMMAND ----------

# MAGIC %pip install mlflow-extensions
# MAGIC %pip install -U mlflow
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeploy
from mlflow_extensions.databricks.prebuilt import prebuilt
import os

os.environ["HF_TOKEN"] = dbutils.secrets.get(
    scope="sri-mlflow-extensions", key="hf-token"
)

deployer = EzDeploy(
  config=prebuilt.text.vllm.COHERE_FOR_AYA_23_35B,
  registered_model_name="main.default.cohere_aya_35b"
)

deployer.download()

deployer.register()

endpoint_name = "sri_cohere_aya"

deployer.deploy(endpoint_name)

# COMMAND ----------

endpoint_name = "sri_cohere_aya"

from mlflow_extensions.serving.compat.openai import OpenAI
from mlflow.utils.databricks_utils import get_databricks_host_creds
from mlflow_extensions.databricks.prebuilt import prebuilt

workspace_host = spark.conf.get("spark.databricks.workspaceUrl")
endpoint_name = f"https://{workspace_host}/serving-endpoints/{endpoint_name}/invocations"
token = get_databricks_host_creds().token

client = OpenAI(
  base_url=endpoint_name,
  api_key=token
)

response = client.chat.completions.create(
    model=prebuilt.text.vllm.COHERE_FOR_AYA_23_35B.engine_config.model,
    messages=[
        {
            "role": "user",
            "content": "Hi what model are you, who trained you?"
        }
    ],
)
response

# COMMAND ----------





import functools
from dataclasses import dataclass, field
from typing import List, Optional

import typing_extensions
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.jobs import GitSource, JobCluster, JobSettings, Task

from mlflow_extensions.databricks.deploy.gpu_configs import (
    ALL_VALID_VM_CONFIGS,
    Cloud,
    GPUConfig,
)
from mlflow_extensions.version import get_mlflow_extensions_version

if typing_extensions.TYPE_CHECKING:
    from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig

EZ_DEPLOY_LITE_PREFIX = "[EZ_DEPLOY_LITE]"
EZ_DEPLOY_TASK = "deployment"
DEFAULT_RUNTIME = "15.4.x-gpu-ml-scala2.12"
DEFAULT_SERVING_NOTEBOOK = (
    "mlflow_extensions/databricks/deploy/ez_deploy_lite_entrypoint"
)


@dataclass(kw_only=True, frozen=True)
class JobsConfig:
    valid_gpus: List[GPUConfig] = field(default_factory=lambda: ALL_VALID_VM_CONFIGS)
    minimum_memory_in_gb: Optional[int] = None

    def smallest_gpu(self, cloud: Cloud) -> GPUConfig:
        filtered_gpus = [gpu for gpu in self.valid_gpus if gpu.cloud == cloud]
        if self.minimum_memory_in_gb is not None:
            filtered_gpus = [
                gpu
                for gpu in filtered_gpus
                if gpu.total_memory_gb > self.minimum_memory_in_gb
            ]
        if len(filtered_gpus) == 0:
            raise ValueError(
                "No valid gpus in cloud please provide the right gpu config"
            )
        return min(filtered_gpus, key=lambda gpu: gpu.total_memory_gb)


def make_base_parameters(
    config: "EzDeployConfig", hf_secret_scope: str, hf_secret_key: str
):
    return {
        "ez_deploy_config": config.serialize_json(),
        "hf_secret_scope": hf_secret_scope or "",
        "hf_secret_key": hf_secret_key or "",
        "pip_reqs": " ".join(config.engine_config.default_pip_reqs()),
    }


def make_cloud_specific_attrs(cloud: Cloud):
    if cloud == Cloud.GCP:
        return {
            "gcp_attributes": {
                "use_preemptible_executors": False,
                "availability": "ON_DEMAND_GCP",
                "zone_id": "HA",
            }
        }
    if cloud == Cloud.AWS:
        return {
            "aws_attributes": {
                "first_on_demand": 1,
                "availability": "SPOT_WITH_FALLBACK",
                "zone_id": "auto",
                "instance_profile_arn": None,
                "spot_bid_price_percent": 100,
            },
        }
    if cloud == Cloud.AZURE:
        return {
            "azure_attributes": {
                "first_on_demand": 1,
                "availability": "ON_DEMAND_AZURE",
                "spot_bid_max_price": -1,
            }
        }
    raise ValueError(f"Cloud {cloud} is not supported")


def make_create_json(
    *,
    job_name: str,
    minimum_memory_in_gb: int,
    cloud_provider: Cloud,
    ez_deploy_config: "EzDeployConfig",
    huggingface_secret_scope: str,
    huggingface_secret_key: str,
    task_name: str = EZ_DEPLOY_TASK,
    runtime: str = DEFAULT_RUNTIME,
    notebook_path: str = DEFAULT_SERVING_NOTEBOOK,
    specific_git_ref: str = None,
):
    vm = JobsConfig(minimum_memory_in_gb=minimum_memory_in_gb).smallest_gpu(
        cloud_provider
    )
    gpu_node = {
        "spark_version": runtime,
        "spark_conf": {
            "spark.master": "local[*, 4]",
            "spark.databricks.cluster.profile": "singleNode",
        },
        "node_type_id": vm.name,
        "driver_node_type_id": vm.name,
        "custom_tags": {"ResourceClass": "SingleNode"},
        "enable_elastic_disk": True,
        "data_security_mode": "NONE",
        "runtime_engine": "STANDARD",
        "num_workers": 0,
    }
    gpu_node.update(make_cloud_specific_attrs(cloud_provider))

    if specific_git_ref:
        parts = specific_git_ref.split("/", maxsplit=1)
        git_type = "git_branch"
        if parts[0].startswith("commit"):
            git_type = "git_commit"
        elif parts[0].startswith("tag"):
            git_type = "git_tag"

        git_source = GitSource.from_dict(
            {
                "git_url": "https://github.com/stikkireddy/mlflow-extensions.git",
                "git_provider": "gitHub",
                git_type: specific_git_ref,
            }
        )
    else:
        git_source = GitSource.from_dict(
            {
                "git_url": "https://github.com/stikkireddy/mlflow-extensions.git",
                "git_provider": "gitHub",
                "git_tag": f"v{get_mlflow_extensions_version()}",
            }
        )

    return {
        "name": job_name,
        "timeout_seconds": 0,
        "max_concurrent_runs": 1,
        "tasks": [
            Task.from_dict(
                {
                    "task_key": task_name,
                    "run_if": "ALL_SUCCESS",
                    "notebook_task": {
                        "notebook_path": notebook_path,
                        "source": "GIT",
                        "base_parameters": make_base_parameters(
                            ez_deploy_config,
                            huggingface_secret_scope,
                            huggingface_secret_key,
                        ),
                    },
                    "job_cluster_key": "deployment_gpu",
                    "timeout_seconds": 0,
                }
            )
        ],
        "git_source": git_source,
        "job_clusters": [
            JobCluster.from_dict(
                {"job_cluster_key": "deployment_gpu", "new_cluster": gpu_node}
            )
        ],
    }


class EzDeployLiteManager:

    def __init__(
        self,
        databricks_host: str = None,
        databricks_token: str = None,
        prefix: str = EZ_DEPLOY_LITE_PREFIX,
    ):
        self._prefix = prefix
        self.databricks_host = databricks_host
        self.databricks_token = databricks_token
        self.client = WorkspaceClient(host=databricks_host, token=databricks_token)

    def make_name(self, name):
        return f"{self._prefix} {name}"

    @functools.lru_cache(maxsize=32)
    def get_jobs(self, job_name):
        return [job for job in self.client.jobs.list(name=job_name)]

    def exists(self, model_deployment_name):
        job_name = self.make_name(model_deployment_name)
        return len(self.get_jobs(job_name)) >= 1

    def make_oai_url(self, cluster_id):
        return f"{self.client.config.host.rstrip('/')}/driver-proxy-api/o/0/{cluster_id}/9989/v1/"

    def get_openai_url(
        self, model_deployment_name, deployment_name: str = EZ_DEPLOY_TASK
    ):
        job_name = self.make_name(model_deployment_name)
        if not self.exists(model_deployment_name):
            raise ValueError(f"Model deployment {model_deployment_name} does not exist")
        job = [job for job in self.client.jobs.list(name=job_name)][0]
        job_id = job.job_id
        for run in self.client.jobs.list_runs(
            active_only=True, expand_tasks=True, job_id=job_id
        ):
            for task in run.tasks:
                if task.task_key == deployment_name:
                    if task.cluster_instance is None:
                        raise ValueError(
                            f"Model deployment {model_deployment_name} has no cluster"
                        )
                    if task.cluster_instance.cluster_id is None:
                        raise ValueError(
                            f"Model deployment {model_deployment_name} has no cluster id"
                        )
                    return self.make_oai_url(task.cluster_instance.cluster_id)
        raise ValueError(
            f"Model deployment {model_deployment_name} has no running cluster endpoint"
        )

    def upsert(
        self,
        model_deployment_name,
        cloud_provider,
        ez_deploy_config: "EzDeployConfig",
        hf_secret_scope=None,
        hf_secret_key=None,
        entrypoint_git_ref: str = None,
    ):
        job_name = self.make_name(model_deployment_name)
        assert (
            ez_deploy_config.serving_config.minimum_memory_in_gb
        ), "Minimum memory in GB is not set"
        create_json = make_create_json(
            job_name=job_name,
            minimum_memory_in_gb=ez_deploy_config.serving_config.minimum_memory_in_gb,
            cloud_provider=cloud_provider,
            ez_deploy_config=ez_deploy_config,
            huggingface_secret_scope=hf_secret_scope,
            huggingface_secret_key=hf_secret_key,
            specific_git_ref=entrypoint_git_ref,
        )
        if self.exists(model_deployment_name):
            job_id = self.get_jobs(job_name)[0].job_id
            reset_data = {"new_settings": JobSettings(**create_json)}
            return self.client.jobs.reset(job_id=job_id, **reset_data)
        self.client.jobs.create(**create_json)

    def start_server(self, model_deployment_name):
        job_name = self.make_name(model_deployment_name)
        if not self.exists(model_deployment_name):
            raise ValueError(f"Model deployment {model_deployment_name} does not exist")
        job = list(self.client.jobs.list(name=job_name))[0]
        job_id = job.job_id
        if len(list(self.client.jobs.list_runs(active_only=True, job_id=job_id))) == 0:
            return self.client.jobs.run_now(job_id=job_id)
        raise ValueError(
            f"Model deployment {model_deployment_name} has an active run, please cancel the run "
            f"and start a new run."
        )



import pytest


@pytest.mark.unit
def test_vllm_config():
    from mlflow_extensions.databricks.prebuilt.text.vllm import NUEXTRACT_CONFIG

    assert len(NUEXTRACT_CONFIG.engine_config.default_pip_reqs()) > 0


@pytest.mark.unit
def test_sglang_config():
    from mlflow_extensions.databricks.prebuilt.text.sglang import GEMMA_2_9B_IT_CONFIG

    pip_reqs = GEMMA_2_9B_IT_CONFIG.engine_config.default_pip_reqs()
    assert len(pip_reqs) > 0, "pip_reqs not found"
    assert (
        "--extra-index-url=https://flashinfer.ai/whl/cu121/torch2.4/" in pip_reqs
    ), "flashinfer extra index url not found"


@pytest.mark.unit
def test_new_library():
    from mlflow_extensions.serving.engines.vllm_engine import VLLMEngineConfig

    cfg = VLLMEngineConfig(model="foo", library_overrides={"torch": "torch==1.9.0"})
    pip_reqs = cfg.default_pip_reqs()
    assert "torch==1.9.0" in pip_reqs, "torch override not found"


@pytest.mark.unit
def test_override_library():
    from mlflow_extensions.serving.engines.vllm_engine import VLLMEngineConfig

    cfg = VLLMEngineConfig(model="foo", library_overrides={"vllm": "vllm=fooversion"})
    pip_reqs = cfg.default_pip_reqs()
    assert "vllm=fooversion" in pip_reqs, "torch override not found"



import json
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union

from mlflow.pyfunc import PythonModelContext

from mlflow_extensions.log import Logger, get_logger
from mlflow_extensions.serving.engines import gpu_utils
from mlflow_extensions.serving.engines.base import Command, EngineConfig, EngineProcess
from mlflow_extensions.serving.engines.huggingface_utils import (
    ensure_chat_template,
    snapshot_download_local,
)
from mlflow_extensions.testing.helper import kill_processes_containing

LOGGER: Logger = get_logger()


@dataclass(frozen=True, kw_only=True)
class SglangEngineConfig(EngineConfig):
    entrypoint_module: str = field(default="sglang.launch_server")
    sglang_command_flags: Dict[str, Optional[str]] = field(default_factory=dict)
    trust_remote_code: bool = field(default=False)  # --trust-remote-code
    context_length: Optional[int] = field(default=None)  # --context-length
    served_model_alias: Optional[str] = field(default=None)  # --served-model-name
    quantization: Optional[str] = field(default=None)  # --quantization
    tokenizer_path: Optional[str] = field(default=None)  # --tokenizer-path
    # generic
    model_artifact_key: str = field(default="model")
    tokenizer_artifact_key: str = field(default="tokenizer")
    verify_chat_template: bool = field(default=True)
    tokenizer_config_file: str = field(default="tokenizer_config.json")
    chat_template_key: str = field(default="chat_template")
    chat_template_builtin_name: str = field(default=None)
    chat_template_json: dict = field(default=None)
    chat_template_file_name: str = field(default="sglang_custom_chat_template.json")

    def _to_sglang_command(self, context: PythonModelContext = None) -> List[str]:
        local_model_path = None
        tokenizer_model_path = None
        if context is not None:
            local_model_path = context.artifacts.get(self.model_artifact_key)
            tokenizer_model_path = context.artifacts.get(self.tokenizer_artifact_key)
        flags = []

        if self.tokenizer_path is not None:
            flags.append("--tokenizer-path")
            flags.append(
                tokenizer_model_path
                if tokenizer_model_path is not None
                else self.tokenizer_path
            )

        # add tensor parallel size flag if we have GPUs
        gpu_count = gpu_utils.get_gpu_count()
        if gpu_count >= 1:
            flags.append("--tensor-parallel-size")
            flags.append(str(gpu_count))

        skip_flags = [
            "--model-path",
            "--context-length",
            "--trust-remote-code",
            "--served-model-name",
            "--tokenizer-path",
            "--quantization",
            "--chat-template",
        ]
        for k, v in self.sglang_command_flags.items():
            if k in skip_flags:
                LOGGER.info(f"Skipping flag {k} use the built in argument")
                continue
            flags.append(k)
            if v is not None:
                flags.append(v)
        if self.trust_remote_code is True:
            flags.append("--trust-remote-code")
        if self.context_length is not None:
            flags.append("--context-length")
            flags.append(str(self.context_length))
        if self.quantization is not None:
            flags.append("--quantization")
            # only valid quantization methods supported
            if self.quantization not in [
                "awq",
                "fp8",
                "gptq",
                "marlin",
                "gptq_marlin",
                "awq_marlin",
                "squeezellm",
                "bitsandbytes",
            ]:
                raise ValueError(f"Invalid quantization {self.quantization}")
            flags.append(self.quantization)
        if self.chat_template_json or self.chat_template_builtin_name:
            flags.append("--chat-template")
            if self.chat_template_builtin_name:
                flags.append(self.chat_template_builtin_name)
            else:
                with open(self.chat_template_file_name, "w") as f:
                    f.write(json.dumps(self.chat_template_json))
                flags.append(self.chat_template_file_name)

        return [
            sys.executable,
            "-m",
            self.entrypoint_module,
            "--host",
            self.host,
            "--port",
            str(self.port),
            "--model-path",
            self.model if local_model_path is None else local_model_path,
            "--served-model-name",
            self.model if self.served_model_alias is None else self.served_model_alias,
            *flags,
        ]

    def _to_run_command(
        self, context: PythonModelContext = None
    ) -> Union[List[str], Command]:
        return self._to_sglang_command(context=context)

    def engine_pip_reqs(
        self,
        *,
        sglang_version: str = "0.3.0",
        flashinfer_extra_index_url: str = "https://flashinfer.ai/whl/cu121/torch2.4/",
        flashinfer_version: str = "0.1.6",
    ) -> Dict[str, str]:
        default_installs = {
            "sglang": f"sglang[all]=={sglang_version}",
        }
        if flashinfer_extra_index_url is not None:
            default_installs["extra_index_url"] = (
                f"--extra-index-url={flashinfer_extra_index_url}"
            )
        default_installs["flashinfer"] = f"flashinfer=={flashinfer_version}"
        return default_installs

    def _setup_snapshot(self, local_dir: str = "/root/models"):
        return snapshot_download_local(repo_id=self.model, local_dir=local_dir)

    def _setup_artifacts(self, local_dir: str = "/root/models"):
        local_path = self._setup_snapshot(local_dir)
        # tokenizer path
        artifacts = {self.model_artifact_key: local_path}
        if self.tokenizer_path is not None and self.model != self.tokenizer_path:
            tokenizer_local_path = snapshot_download_local(
                repo_id=self.tokenizer_path, local_dir=local_dir, tokenizer_only=True
            )
            artifacts[self.tokenizer_artifact_key] = tokenizer_local_path
        return artifacts

    def _verify_chat_template(self, artifacts: Dict[str, str]):
        if self.tokenizer_path is None:
            model_dir_path = Path(artifacts[self.model_artifact_key])
        else:
            model_dir_path = Path(artifacts[self.tokenizer_artifact_key])
        tokenizer_config_file = model_dir_path / self.tokenizer_config_file
        ensure_chat_template(
            tokenizer_file=str(tokenizer_config_file),
            chat_template_key=self.chat_template_key,
        )

    def setup_artifacts(self, local_dir: str = "/root/models"):
        artifacts = self._setup_artifacts(local_dir)
        if self.verify_chat_template is True:
            self._verify_chat_template(artifacts)
        return artifacts

    def supported_model_architectures(self) -> List[str]:
        return [
            "MiniCPMForCausalLM",
            "StableLmForCausalLM",
            "LlavaVidForCausalLM",
            "Qwen2MoeForCausalLM",
            "Qwen2ForCausalLM",
            "GPTBigCodeForCausalLM",
            "ChatGLMForCausalLM",
            "ChatGLMModel",
            "DbrxForCausalLM",
            "LlavaLlamaForCausalLM",
            "LlavaQwenForCausalLM",
            "LlavaMistralForCausalLM",
            "InternLM2ForCausalLM",
            "QuantMixtralForCausalLM",
            "DeepseekV2ForCausalLM",
            "CohereForCausalLM",
            "MixtralForCausalLM",
            "YiVLForCausalLM",
            "LlamaEmbeddingModel",
            "MistralModel",
            "LlamaForClassification",
            "QWenLMHeadModel",
            "Gemma2ForCausalLM",
            "DeepseekForCausalLM",
            "LlamaForCausalLM",
            "GemmaForCausalLM",
            "MistralForCausalLM",
            "Grok1ModelForCausalLM",
        ]


class SglangEngineProcess(EngineProcess):

    @property
    def engine_name(self) -> str:
        return "sglang-engine"

    def health_check(self) -> bool:
        try:
            resp = self.server_http_client.get("/health")
            return resp.status_code == 200
        except Exception as e:
            LOGGER.info(
                f"Health check failed with error {e}; server may not be up yet or crashed;"
            )
            return False

    def cleanup(self) -> None:
        try:
            import ray

            ray.shutdown()
        except Exception:
            pass

        try:
            kill_processes_containing("sglang")
            # sglang uses vllm under the hood and may spawn vllm proc
            kill_processes_containing("vllm")
            kill_processes_containing("ray")
            kill_processes_containing("from multiprocessing")
        except Exception:
            pass



import typing

from mlflow_extensions.testing.helper import Modality, inject_openai_client, run_if

if typing.TYPE_CHECKING is True:
    from openai import OpenAi

    from mlflow_extensions.testing.runner import ModelContextRunner


@run_if(modality=Modality.VISION.value)
@inject_openai_client
def query_vision(
    *,
    ctx: "ModelContextRunner",
    client: "OpenAi",
    model: str,
    modality_type: str = None,
    host: str = "0.0.0.0",  # noqa
    port: int = 9989,  # noqa
    repeat_n: int = 5,
):
    count = 0
    # repeat a few times
    while count < repeat_n:
        from pydantic import BaseModel

        class ExpectedJson(BaseModel):
            outside: bool
            inside: bool
            boardwalk: bool
            grass: bool

        try:
            response = client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": "Describe the images"},
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": "https://mir-s3-cdn-cf.behance.net/project_modules/max_1200/73fbe271026179.5bb6e7af358b6.jpg"
                                },
                            },
                        ],
                    }
                ],
            )
            ctx.add_success(result=response.choices[0].message.content.strip())

            response = client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": "Is the image indoors or outdoors?",
                            },
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
                                },
                            },
                        ],
                    }
                ],
                extra_body={"guided_json": ExpectedJson.schema()},
            )
            ctx.add_success(result=response.choices[0].message.content.strip())
        except Exception as e:
            print(e)
            ctx.add_error(error_msg=str(e))
        finally:
            count += 1


@run_if(modality=Modality.VISION.value)
@inject_openai_client
def query_vision_multi_input(
    *,
    ctx: "ModelContextRunner",
    client: "OpenAi",
    model: str,
    modality_type: str = None,
    host: str = "0.0.0.0",  # noqa
    port: int = 9989,  # noqa
    repeat_n: int = 5,
):
    count = 0
    # repeat a few times
    while count < repeat_n:
        try:
            response = client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": "Compare the differences between the images?",
                            },
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
                                },
                            },
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": "https://mir-s3-cdn-cf.behance.net/project_modules/max_1200/73fbe271026179.5bb6e7af358b6.jpg"
                                },
                            },
                        ],
                    }
                ],
            )
            ctx.add_success(result=response.choices[0].message.content.strip())
        except Exception as e:
            print(e)
            ctx.add_error(error_msg=str(e))
        finally:
            count += 1



# Databricks notebook source
dbutils.widgets.text("ez_deploy_config", "")
dbutils.widgets.text("hf_secret_scope", "")
dbutils.widgets.text("hf_secret_key", "")
dbutils.widgets.text("pip_reqs", "")

# COMMAND ----------

ez_deploy_config = dbutils.widgets.get("ez_deploy_config")
hf_secret_scope = dbutils.widgets.get("hf_secret_scope")
hf_secret_key = dbutils.widgets.get("hf_secret_key")
pip_reqs = dbutils.widgets.get("pip_reqs")
assert ez_deploy_config, "ez_deploy_config is required"
assert pip_reqs, "pip_reqs is required"

# COMMAND ----------

from IPython import get_ipython

ipython = get_ipython()
ipython.run_line_magic("pip", f"install {pip_reqs} hf_transfer")
ipython.run_line_magic("pip", f"install -U openai")
dbutils.library.restartPython()

# COMMAND ----------

ez_deploy_config = dbutils.widgets.get("ez_deploy_config")
hf_secret_scope = dbutils.widgets.get("hf_secret_scope")
hf_secret_key = dbutils.widgets.get("hf_secret_key")
pip_reqs = dbutils.widgets.get("pip_reqs")
assert ez_deploy_config, "ez_deploy_config is required"
assert pip_reqs, "pip_reqs is required"

# COMMAND ----------

import os

os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"
if hf_secret_scope and hf_secret_key:
    os.environ["HF_TOKEN"] = dbutils.secrets.get(
        scope=hf_secret_scope, key=hf_secret_key
    )

# COMMAND ----------

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig

config = EzDeployConfig.from_json(ez_deploy_config)

# COMMAND ----------

engine_process = config.to_proc()

# COMMAND ----------

artifacts = config.download_artifacts()

# COMMAND ----------

from mlflow.pyfunc import PythonModelContext

ctx = PythonModelContext(artifacts=artifacts, model_config={})

# COMMAND ----------

import ray

from mlflow_extensions.testing.helper import kill_processes_containing

ray.shutdown()
kill_processes_containing("vllm")
kill_processes_containing("ray")
kill_processes_containing("sglang")
kill_processes_containing("from multiprocessing")
engine_process.start_proc(ctx, health_check_thread=False)

# COMMAND ----------

import time

while True:
    time.sleep(1)



from dataclasses import dataclass, field

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig, ServingConfig
from mlflow_extensions.serving.engines import VLLMEngineConfig, VLLMEngineProcess

_ENGINE = VLLMEngineProcess
_ENGINE_CONFIG = VLLMEngineConfig

# https://huggingface.co/numind/NuExtract
NUEXTRACT_CONFIG = EzDeployConfig(
    name="nuextract_small",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="numind/NuExtract",
        trust_remote_code=True,
        guided_decoding_backend="outlines",
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=20,
    ),
)

# https://huggingface.co/numind/NuExtract-tiny
NUEXTRACT_TINY_CONFIG = EzDeployConfig(
    name="nuextract_tiny",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="numind/NuExtract-tiny",
        trust_remote_code=True,
        guided_decoding_backend="outlines",
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=10,
    ),
)

# https://huggingface.co/NousResearch/Hermes-3-Llama-3.1-8B
NOUS_HERMES_3_LLAMA_3_1_8B_64K = EzDeployConfig(
    name="hermes_3_llama_3_1_8b_64k",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="NousResearch/Hermes-3-Llama-3.1-8B",
        trust_remote_code=True,
        guided_decoding_backend="outlines",
        max_model_len=64000,
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=60,
    ),
)


# https://huggingface.co/NousResearch/Hermes-3-Llama-3.1-8B
NOUS_HERMES_3_LLAMA_3_1_8B_128K = EzDeployConfig(
    name="hermes_3_llama_3_1_8b_128k",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="NousResearch/Hermes-3-Llama-3.1-8B",
        trust_remote_code=True,
        guided_decoding_backend="outlines",
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=110,
    ),
)

COHERE_FOR_AYA_23_35B = EzDeployConfig(
    name="cohere_aya_35b",
    engine_proc=_ENGINE,
    engine_config=_ENGINE_CONFIG(
        model="CohereForAI/aya-23-35B",
        guided_decoding_backend="outlines",
        vllm_command_flags={
            "--max-num-seqs": 128,
            "--gpu-memory-utilization": 0.98,
            "--distributed-executor-backend": "ray",
        },
    ),
    serving_config=ServingConfig(
        # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
        minimum_memory_in_gb=76,
    ),
)


# https://huggingface.co/microsoft/Phi-3.5-MoE-instruct
# PHI_3_5_MOE_INSTRUCT_8K = EzDeployConfig(
#     name="phi_3_5_moe_instruct_8k",
#     engine_proc=_ENGINE,
#     engine_config=_ENGINE_CONFIG(
#         model="microsoft/Phi-3.5-MoE-instruct",
#         trust_remote_code=True,
#         max_model_len=8192,
#         guided_decoding_backend="outlines"
#     ),
#     serving_config=ServingConfig(
#         # rough estimate for the engines this includes model weights + kv cache + overhead + intermediate states
#         minimum_memory_in_gb=50,
#     ),
# )


@dataclass(frozen=True)
class VllmText:
    NUEXTRACT: EzDeployConfig = field(default_factory=lambda: NUEXTRACT_CONFIG)
    NUEXTRACT_TINY: EzDeployConfig = field(
        default_factory=lambda: NUEXTRACT_TINY_CONFIG
    )
    NOUS_HERMES_3_LLAMA_3_1_8B_64K: EzDeployConfig = field(
        default_factory=lambda: NOUS_HERMES_3_LLAMA_3_1_8B_64K
    )
    NOUS_HERMES_3_LLAMA_3_1_8B_128K: EzDeployConfig = field(
        default_factory=lambda: NOUS_HERMES_3_LLAMA_3_1_8B_128K
    )
    COHERE_FOR_AYA_23_35B: EzDeployConfig = field(
        default_factory=lambda: COHERE_FOR_AYA_23_35B
    )
    # PHI_3_5_MOE_INSTRUCT_8K = PHI_3_5_MOE_INSTRUCT_8K



from dataclasses import dataclass, field

from mlflow_extensions.databricks.prebuilt.audio.vllm import VllmAudio


@dataclass(frozen=True)
class AudioLM:

    vllm: VllmAudio = field(default_factory=lambda: VllmAudio())



import pandas as pd

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig
from mlflow_extensions.databricks.deploy.gpu_configs import Cloud
from mlflow_extensions.databricks.prebuilt import prebuilt
from mlflow_extensions.serving.engines import VLLMEngineConfig

try:
    import tabulate
except ImportError as e:
    print("Please install tabulate to run this script. pip install tabulate")
    raise e

if __name__ == "__main__":
    data = []
    for model_type, model_arg in prebuilt.__dict__.items():
        for serving_type, serving_framework_cfg in model_arg.__dict__.items():
            for key, value in serving_framework_cfg.__dict__.items():
                value: EzDeployConfig
                context_length = "Default"
                if isinstance(value.engine_config, VLLMEngineConfig):
                    if value.engine_config.max_model_len is not None:
                        context_length = str(value.engine_config.max_model_len)
                azure_gpu_cfg = value.serving_config.smallest_gpu(Cloud.AZURE)
                aws_gpu_cfg = value.serving_config.smallest_gpu(Cloud.AWS)
                data.append(
                    {
                        "model_type": model_type,
                        "cfg_path": f"prebuilt.{model_type}.{serving_type}.{key}",
                        "huggingface_link": f"https://huggingface.co/{value.engine_config.model}",
                        "context_length": context_length,
                        "min_azure_ep_type_gpu": f"{azure_gpu_cfg.name} [{azure_gpu_cfg.gpu_type.value}x{azure_gpu_cfg.gpu_count} {azure_gpu_cfg.total_memory_gb}GB]",
                        "min_aws_ep_type_gpu": f"{aws_gpu_cfg.name} [{aws_gpu_cfg.gpu_type.value}x{aws_gpu_cfg.gpu_count} {aws_gpu_cfg.total_memory_gb}GB]",
                    }
                )

    df = pd.DataFrame(data)
    print(df.to_markdown(index=False))



import fnmatch
import subprocess
import time
from dataclasses import asdict, dataclass
from typing import List, Optional, Tuple

from mlflow_extensions.databricks.deploy.ez_deploy import EzDeployConfig
from mlflow_extensions.databricks.deploy.gpu_configs import GPUConfig
from mlflow_extensions.databricks.prebuilt import prebuilt
from mlflow_extensions.serving.engines import VLLMEngineConfig
from mlflow_extensions.serving.engines.base import EngineProcess
from mlflow_extensions.testing.audio_basic import (
    encode_audio_base64_from_url,
    query_audio,
)
from mlflow_extensions.testing.helper import ServerFramework, kill_processes_containing
from mlflow_extensions.testing.text_basic import query_text
from mlflow_extensions.testing.vision_basic import (
    query_vision,
    query_vision_multi_input,
)


def make_process_and_get_artifacts(
    config: EzDeployConfig, local_dir=None
) -> Tuple[EngineProcess, dict]:
    if local_dir is not None:
        artifacts = config.engine_config.setup_artifacts(local_dir)
    else:
        artifacts = config.engine_config.setup_artifacts()
    engine = config.engine_proc(config=config.engine_config)
    return engine, artifacts


@dataclass
class RequestResult:
    model: str
    config_name: str
    framework: str
    gpu: str
    output: str
    error_msg: str
    is_error: bool
    did_server_crash: bool
    cloud: str
    server_command: str

    @staticmethod
    def make_df_friendly(results: List["RequestResult"]):
        return [asdict(result) for result in results]


class ModelContextRunner:
    def __init__(
        self,
        ez_config: EzDeployConfig,
        current_gpu: GPUConfig,
        skip_health_check: bool = True,
    ):
        self.ez_config = ez_config
        self.engine: Optional[EngineProcess] = None
        self.artifacts: Optional[dict] = None
        self.model_context = None
        self._results: List[RequestResult] = []
        self.current_gpu = current_gpu
        self.command: Optional[str] = None
        self._skip_health_check = skip_health_check

    def __enter__(self):
        try:
            from mlflow.pyfunc import PythonModelContext
        except ImportError:
            raise ImportError(
                "mlflow is required to use this class run pip install -U mlflow"
            )
        self.engine, self.artifacts = make_process_and_get_artifacts(self.ez_config)
        self.add_success(result="SUCCESSFULLY LOADED ARTIFACTS")
        self.model_context = PythonModelContext(
            artifacts=self.artifacts, model_config={}
        )
        self.ez_config.engine_config.to_run_command(self.model_context)
        self.engine.start_proc(
            self.model_context, health_check_thread=not self._skip_health_check
        )
        self.add_success(result="SUCCESSFULLY STARTED SERVER")
        return self

    def add_error(self, *, error_msg: str):
        self.results.append(
            RequestResult(
                model=self.ez_config.engine_config.model,
                config_name=self.ez_config.name,
                gpu=f"{self.current_gpu.name}x{self.current_gpu.gpu_count}",
                output="",
                error_msg=error_msg,
                is_error=True,
                cloud=self.current_gpu.cloud.value,
                did_server_crash=self.engine.is_process_healthy() is False,
                framework=self.ez_config.engine_proc.__name__,
                server_command=self.command,
            )
        )

    def add_success(self, *, result: str):
        self.results.append(
            RequestResult(
                model=self.ez_config.engine_config.model,
                config_name=self.ez_config.name,
                gpu=f"{self.current_gpu.name}x{self.current_gpu.gpu_count}",
                error_msg="",
                output=result,
                is_error=False,
                cloud=self.current_gpu.cloud.value,
                did_server_crash=self.engine.is_process_healthy() is False,
                framework=self.ez_config.engine_proc.__name__,
                server_command=self.command,
            )
        )

    @property
    def results(self):
        return self._results

    @property
    def results_as_dict(self):
        return [asdict(result) for result in self._results]

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.add_error(error_msg=str(exc_val))

        if self.engine is not None and self.engine.server_process is not None:
            self.engine.stop_proc()


def run_all_tests(
    *,
    gpu_config: GPUConfig,
    server_framework: ServerFramework,
    model_filter_fnmatch_str: str = "*",
    ezconfig_name_filter_fnmatch_str: str = "*",
) -> List[RequestResult]:
    # gettysburg.wav is a 17 second audio file
    audio_data = encode_audio_base64_from_url(
        "https://www2.cs.uic.edu/~i101/SoundFiles/gettysburg.wav"
    )
    results = []

    for modality, model_arg in prebuilt.__dict__.items():
        for serving_type, serving_framework_cfg in model_arg.__dict__.items():
            for key, ezconfig in serving_framework_cfg.__dict__.items():
                ezconfig: EzDeployConfig
                if serving_type != server_framework.value:
                    continue
                if (
                    ezconfig.serving_config.minimum_memory_in_gb
                    > gpu_config.total_memory_gb
                ):
                    continue
                if (
                    fnmatch.fnmatch(
                        ezconfig.engine_config.model, model_filter_fnmatch_str
                    )
                    is False
                ):
                    print(
                        f"SKIPPING {ezconfig.engine_config.model} because it does not match "
                        f"{model_filter_fnmatch_str}"
                    )
                    continue
                if (
                    fnmatch.fnmatch(ezconfig.name, ezconfig_name_filter_fnmatch_str)
                    is False
                ):
                    print(
                        f"SKIPPING {ezconfig.engine_config.model} because it does not match "
                        f"{ezconfig_name_filter_fnmatch_str}"
                    )
                    continue

                # kill everything before starting a new process
                # we want to make sure ports are available
                kill_processes_containing("sglang.launch_server")
                kill_processes_containing("vllm.entrypoints.openai.api_server")
                with ModelContextRunner(
                    ez_config=ezconfig, current_gpu=gpu_config
                ) as ctx:
                    query_audio(
                        ctx=ctx,
                        model=ezconfig.engine_config.model,
                        modality_type=modality,
                        audio_data=audio_data,
                    )
                    query_vision(
                        ctx=ctx,
                        model=ezconfig.engine_config.model,
                        modality_type=modality,
                    )
                    query_text(
                        ctx=ctx,
                        model=ezconfig.engine_config.model,
                        modality_type=modality,
                    )
                    if (
                        server_framework == ServerFramework.VLLM
                        and isinstance(ezconfig.engine_config, VLLMEngineConfig)
                        and ezconfig.engine_config.max_num_images is not None
                        and ezconfig.engine_config.max_num_images > 1
                    ):
                        # TODO: accept max num images to ensure we test multiple images
                        query_vision_multi_input(
                            ctx=ctx,
                            model=ezconfig.engine_config.model,
                            modality_type=modality,
                        )

                results.extend(ctx.results)
                print("WAITING FOR SERVER TO CLEANLY CLOSE BEFORE STARTING NEW SERVER")
                time.sleep(5)

    return results



# Databricks notebook source
# MAGIC %pip install -U "mlflow-extensions>=0.9.0,<1.0.0"
# MAGIC dbutils.library.restartPython()

# COMMAND ----------

import mlflow.pyfunc

class CustomModel(mlflow.pyfunc.PythonModel):

    def load_context(self, context):
        # raise ValueError("some bug getting artifacts or importing stuff")
        return

    def predict(self, context, model_input, params=None):
        return model_input*2

# COMMAND ----------

with mlflow.start_run() as run:
    model = CustomModel()
    mlflow.pyfunc.log_model(
        "model",
        python_model=model,
    )

# COMMAND ----------

run_uri = f"runs:/{run.info.run_id}/model"
run_uri

# COMMAND ----------

import numpy as np
loaded_model = mlflow.pyfunc.load_model(run_uri)
# context, model_input, params (default None)
first_attempt = loaded_model.predict([[1, 2, 3]])
second_attempt = loaded_model.predict(np.array([1, 2, 3]))
first_attempt, second_attempt

# COMMAND ----------

from mlflow_extensions.testing.fixures import LocalTestServer
from mlflow.utils.databricks_utils import get_databricks_host_creds

server_configs = {
  "model_uri": run_uri,
  "registry_host": get_databricks_host_creds().host,
  "registry_token": get_databricks_host_creds().token,
  "use_local_env": True
}

with LocalTestServer(**server_configs) as server:
    resp = server.query(payload={
      "inputs": [1, 2, 3]
    }).json()
    print(resp)
    assert resp == {'predictions': [2, 4, 6]}, "Predictions should be double of the input"

# COMMAND ----------

from mlflow_extensions.testing.fixures import LocalTestServer
from mlflow.utils.databricks_utils import get_databricks_host_creds

server_configs = {
  "model_uri": run_uri,
  "registry_host": get_databricks_host_creds().host,
  "registry_token": get_databricks_host_creds().token,
  "use_local_env": True
}

server = LocalTestServer(**server_configs)

server.start()

server.wait_and_assert_healthy()

resp = server.query(payload={
  "inputs": [1, 2, 3]
}).json()

print(resp)

assert resp == {'predictions': [2, 4, 6]}, "Predictions should be double of the input"

server.stop()
